package metaf2xml::FGFS;

# copyright (c) metaf2xml 2006
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

use strict;
use warnings;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
require Exporter;

use POSIX qw(mktime strftime);

# path may be changed by install
use lib '/opt/metaf2xml/lib'; # METAF2XML_LIB
use metaf2xml::basic;

@ISA = qw(Exporter);

@EXPORT = qw(
    printMetar_FGFS
);

my $strict_fgfs;

my $STATUTE_MILE2METER = 1609.3412196;
my $NM2KM              = 1.852;
my $METER_PER_SEC2KT   = (3.6 / $NM2KM);

# see http://www.srh.noaa.gov/eyw/HTML/Jim_Clouds/CloudTypes/low.html
my %cloud_types_low = (
    '0' => 'no clouds',
    '1' => $cloud_types{CU} . ' (little vertical extent)',
    '2' => $cloud_types{CU} . ' (moderate/strong vertical extent or '
           . $cloud_types{TCU} . ')',
    '3' => $cloud_types{CB} . ' (tops not fibrous or in form of an anvil)',
    '4' => $cloud_types{SC} . ' (formed by spreading '
           . $cloud_types{CU} . ')',
    '5' => $cloud_types{SC} . ' (not formed by spreading '
           . $cloud_types{CU} . ')',
    '6' => $cloud_types{ST} . ' (fairly continuous sheet and/or layer)',
    '7' => $cloud_types{SF} . ' or ' . $cloud_types{CF} . ' of bad weather',
    '8' => $cloud_types{CU} . ' and ' . $cloud_types{SC}
           . ' (not formed by spreading ' . $cloud_types{CU} . ')',
    '9' => $cloud_types{CB} . ' (tops fibrous or in form of an anvil)',
);

my %cloud_types_middle = (
    '0' => 'no clouds',
    '1' => $cloud_types{AS} . ' (predominantly semi-transparent)',
    '2' => $cloud_types{AS} . ' (predominantly opaque)',
    '3' => $cloud_types{AC} . ' (not progressively invading the sky)',
    '4' => $cloud_types{AC} . ' (lenticularis)',
    '5' => $cloud_types{AC}
           . ' (progressively invading the sky or semi-transparent)',
    '6' => $cloud_types{AC} . ' (formed by spreading '
           . $cloud_types{CU} . ' or ' . $cloud_types{CB} . ')',
    '7' => $cloud_types{AC} . ' (mainly opaque, not expanding; or with '
           . $cloud_types{AS} . ')',
    '8' => $cloud_types{AC} . ' (with sproutings or tufts)',
    '9' => 'chaotic sky with ' . $cloud_types{AC} . ' at several levels',
);

my %cloud_types_high = (
    '0' => 'no clouds',
    '1' => $cloud_types{CI} . ' (filaments, strands, or hooks)',
    '2' => $cloud_types{CI} . ' (dense, with sproutings, or in tufts)',
    '3' => $cloud_types{CI} . ' (dense, remaining from cumulonimbus anvil)',
    '4' => $cloud_types{CI}
           . ' (hooks or filaments, progressively invading the sky)',
    '5' => $cloud_types{CI}
           . ' and/or ' . $cloud_types{SC}
           . ' (progressively invading the sky, veil <45 above horizon)',
    '6' => $cloud_types{CI}
           . ' and/or ' . $cloud_types{SC}
           . ' (progressively invading the sky, veil >45 above horizon)',
    '7' => $cloud_types{CS} . ' (covering the entire sky)',
    '8' => $cloud_types{CS} . ' (not invading or covering the entire sky)',
    '9' => $cloud_types{CC} . ' (as predominant high cloud)',
);

sub _dir2compass8 {
    my $dir = shift;
    my ($compass, $ii);

    $ii = sprintf "%.0f", $dir / 45; # with rounding
    while ($ii < 0) {
        $ii += 8;
    }
    while ($ii > 7) {
        $ii -= 8;
    }
    return ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']->[$ii];
}

sub _dir2compass16 {
    my $dir = shift;
    my ($compass, $ii);

    $ii = sprintf "%.0f", $dir / 45 * 2; # with rounding
    while ($ii < 0) {
        $ii += 16;
    }
    while ($ii > 15) {
        $ii -= 16;
    }
    return ['N', 'NNE', 'NE', 'ENE', 'E', 'ESE', 'SE', 'SSE',
            'S', 'SSW', 'SW', 'WSW', 'W', 'WNW', 'NW', 'NNW']->[$ii];
}

sub _printVis_FGFS {
    my ($v_na, $vis, $unitThreshold) = @_;
    my ($msg, $gt_lt, $amountM);

    return "(invalid format: '$vis->{invalidFormat}')"
        if exists $vis->{invalidFormat};
    return $v_na if exists $vis->{notAvailable};
    $gt_lt = '';
    $gt_lt = '>=' if exists $vis->{isEqualGreater};
    $gt_lt = '<' if exists $vis->{isLess};
    if (exists $vis->{unitLength}) {
        if ($vis->{unitLength} eq 'M') {
            $amountM = $vis->{distance};
            if ($amountM == 0) {
                $amountM = 50;
                $gt_lt = '<';
            }
            $gt_lt  = '>=' if $amountM == 9999;
        } elsif ($vis->{unitLength} eq 'KM') {
            $amountM = $vis->{distance} * 1000;
        } elsif ($vis->{unitLength} eq 'SM') {
            $amountM = $vis->{distance} * $STATUTE_MILE2METER
        }
    } else {
        $amountM = $vis->{distance} * 100 * $FT2M;
    }
    $msg = $gt_lt;
    if ($unitThreshold && $amountM >= $unitThreshold) {
        $msg .= sprintf "%g km", rnd($amountM / 1000, 0.1);
    } else {
        $msg .= sprintf "%.0f m", rnd($amountM, 10);
    }
    $msg .= ' ' . $vis->{compassDir} if exists $vis->{compassDir};
    $msg .= sprintf "\t\t\t\t\t$gt_lt%g US-miles",
            rnd($amountM / $STATUTE_MILE2METER, 0.1);
    $msg .= ' ' . $vis->{compassDir} if exists $vis->{compassDir};
    $msg .= "\n\t\t\t(No Directional Variation reporting capabilty)"
        if exists $vis->{NDV};
    return $msg;
}

# out: (is_greater, kt, km/h, mph, m/s), all rounded to 1/10
sub _mkWindSpeedVal {
    my ($speed, $speedGreater) = @_;

    return ($speedGreater,
            rnd($speed, 0.1),
            rnd($speed * $NM2KM, 0.1),
            rnd($speed * $NM2KM / $STATUTE_MILE2METER * 1000, 0.1),
            rnd($speed * $NM2KM / 3.6, 0.1));
}

# in: dir, speed, speedGreater, gustSpeed, gustSpeedGreater,
#     windVarLeft, windVarRight
#     dir+speed=/////:  return $w_na   ->[0]
#     dir+speed=00000:  return $w_calm ->[0]
#     dir=///:          $w_dir_na                                 ->[1]
#     dir=VRB:          $w_dir_vrb                                ->[1]
#     dir=\d+:          sprintf $w_dir (dir, compass8, compass16) ->[1]
#     speed=//:         $w_sp_na                                          ->[2]
#     speed=\d+:        sprintf $fmt_wind (speed_gt, kt, km/h, mp/h, m/s) ->[2]
#     gust=\d+:         sprintf $fmt_gust (gust_gt, kt, km/h, mph, m/s) ->[3]
#     windVarLeft:      sprintf $fmt_var 2x(dir, compass8, compass16)  ->[4]
sub _mkWindVal {
    my ($w, $speedGreater, $gustSpeedGreater,
        $w_na, $w_calm, $w_dir_na, $w_dir_vrb, $w_dir, $w_sp_na,
        $fmt_wind, $fmt_gust, $fmt_var) = @_;
    my (@ret, $speedKT);

    return ($w_na, 0, 0, 0, 0) if exists $w->{notAvailable};

    if (  exists $w->{dir} && $w->{dir} == 0
        && exists $w->{speed} && $w->{speed} == 0
        && !exists $w->{gustSpeed})
    {
        $ret[0] = $w_calm;
        $ret[1] = '';
        $ret[2] = '';
    } else {
        $ret[0] = '';

        if (exists $w->{dirNotAvailable}) {
            $ret[1] = $w_dir_na;
        } elsif (exists $w->{dirVariable}) {
            $ret[1] = $w_dir_vrb;
        } else {
            $ret[1] = sprintf $w_dir, $w->{dir}, _dir2compass8($w->{dir}),
                                                 _dir2compass16($w->{dir});
        }

        if (exists $w->{speedNotAvailable}) {
            $ret[2] = $w_sp_na;
        } else {
            if ($w->{unitSpeed} eq 'KT') {
                $speedKT = $w->{speed} + 0;
            } elsif ($w->{unitSpeed} eq 'KMH') {
                $speedKT = $w->{speed} / $NM2KM;
            } else {
                $speedKT = $w->{speed} * $METER_PER_SEC2KT;
            }
            $ret[2] = sprintf $fmt_wind, _mkWindSpeedVal $speedKT,
                              $speedGreater;
        }
    }

    $ret[3] = '';
    if (exists $w->{gustSpeed}) {
        if ($w->{unitSpeed} eq 'KT') {
            $speedKT = $w->{gustSpeed} + 0;
        } elsif ($w->{unitSpeed} eq 'KMH') {
            $speedKT = $w->{gustSpeed} / $NM2KM;
        } else {
            $speedKT = $w->{gustSpeed} * $METER_PER_SEC2KT;
        }
        $ret[3] = sprintf $fmt_gust, _mkWindSpeedVal $speedKT,
                          $gustSpeedGreater;
    }

    $ret[4] = '';
    if (exists $w->{windVarLeft}) {
        $ret[4] = sprintf $fmt_var, $w->{windVarLeft},
                                    _dir2compass8($w->{windVarLeft}),
                                    _dir2compass16($w->{windVarLeft}),
                                    $w->{windVarRight},
                                    _dir2compass8($w->{windVarRight}),
                                    _dir2compass16($w->{windVarRight});
    }

    return @ret;
}

sub _printWind_FGFS {
    my $w = shift;
    my (@ret, $fmt1, $fmt2);

    $fmt1 = ' at %1$s%3$g km/h';
    $fmt2 = '%1$s%2$g kt = %1$s%4$g mph = %1$s%5$g m/s';
    return "(not available)\n"
        if exists $w->{notAvailable};
    return "(invalid format: '$w->{invalidFormat}')\n"
        if exists $w->{invalidFormat};
    @ret = _mkWindVal $w, $w->{speedGreater} ? '>' : '',
                      $w->{gustSpeedGreater} ? '>' : '',
                      '(not available)',
                      'calm',
                      'direction not available,',
                      'from variable directions',
                      'from the %3$s (%1$d)',
                      ', speed not available',
                      "$fmt1\t\t$fmt2",
                      "\n\t\t\twith gusts$fmt1\t\t\t$fmt2",
                      "\n\t\t\t" . 'variable from %3$s to %6$s (%1$d--%4$d)';
    return $ret[0] . $ret[1] . $ret[2] . $ret[3] . $ret[4] . "\n";
}

sub _printCloudCover {
    return {  SKC => $strict_fgfs ? 'clear skies' : 'sky clear',
              NSC => $strict_fgfs ? 'clear skies' : 'nil significant clouds',
              CLR => $strict_fgfs ? 'clear skies' : 'no clouds below 10000 ft',
              NCD => 'no cloud detected',
             '///'=> '(not available)',
              FEW => 'few clouds',
              SCT => 'scattered clouds',
              BKN => 'broken clouds',
              OVC => 'sky overcast' }->{(shift)};
}

sub _printCloud {
    my ($metar) = @_;
    my ($hdr, $msg, $ii);

    for ($strict_fgfs ? () : @{$metar->{cloud}}) {
        if (exists $_->{isCeiling}) {
            $msg .= "ceiling:\t\tat " . $_->{cloudBase} * 100 . ' ft';
            $msg .=   "\t\t\t\t"
                    . rnd($_->{cloudBase} * 100 * $FT2M, 10) . " m\n";
            last;
        }
    }

    $hdr .= "Sky condition:\t\t";
    for (@{$metar->{cloud}}) {
        $msg .= $hdr;
        $hdr = "\t\t\t";
        if (exists $_->{notAvailable}) {
            $msg .= '(not available)';
        } elsif (exists $_->{invalidFormat}) {
            $msg .= "(invalid format: '$_->{invalidFormat}')";
        } elsif (exists $_->{noClouds}) {
            $msg .= _printCloudCover $_->{noClouds};
        } else {
            $msg .= _printCloudCover $_->{cloudCover};
        }
        if (exists $_->{cloudBase}) {
            $msg .= ' at ' . $_->{cloudBase} * 100 . ' ft'
        } elsif (exists $_->{cloudBaseNotAvailable}) {
            $msg .= ' (cloud base not observable)';
        }
        if (exists $_->{cloudTypeNotAvailable}) {
            $msg .= ' (cloud type not observable)';
        } elsif (exists $_->{cloudType}) {
            $msg .= ' (' . $cloud_types{$_->{cloudType}} . ')';
        }
        $msg .=   "\t\t\t"
                . rnd($_->{cloudBase} * 100 * $FT2M, 10) . " m"
            if exists $_->{cloudBase};
        $msg .= "\n";
    }

    $msg .= "Worst cloud cover:\t"
            . _printCloudCover($metar->{cloudMaxCover}) . "\n"
        if !$strict_fgfs && exists $metar->{cloudMaxCover};

    return $msg;
}

sub _mkTempVal {
    my ($t, $t_na, $fmt_t, $d_na, $fmt_d, $fmt_rh) = @_;
    my (@ret, $sign, $temp, $dew);

    @ret = ('', '', '');
    if (exists $t->{air}{temp}) {
        if ($t->{air}{tempUnit} eq 'C') {
            $ret[0] = sprintf $fmt_t, $t->{air}{temp},
                                      rnd($t->{air}{temp} * 1.8 + 32, 0.1);
        } else {
            $ret[0] = sprintf $fmt_t, rnd(($t->{air}{temp} - 32) / 1.8, 0.1),
                                      $t->{air}{temp};
        }
    } else {
        $ret[0] = $t_na;
    }
    if (exists $t->{dewpoint}) {
        if (exists $t->{dewpoint}{temp}) {
            if ($t->{dewpoint}{tempUnit} eq 'C') {
                $ret[1] = sprintf $fmt_d, $t->{dewpoint}{temp},
                                  rnd($t->{dewpoint}{temp} * 1.8 + 32, 0.1);
            } else {
                $ret[1] = sprintf $fmt_d,
                                  rnd(($t->{dewpoint}{temp} - 32) / 1.8, 0.1),
                                  $t->{dewpoint}{temp} - 32;
            }
        } else {
            $ret[1] = $d_na;
        }
    }
    $ret[2] = sprintf $fmt_rh, $t->{relHumid1}
        if exists $t->{relHumid1};
    return @ret;
}

sub _printTemp_FGFS {
    my ($t, $hdr1, $hdr2) = @_;
    $hdr1 = "Temperature:\t\t" unless $hdr1;
    $hdr2 = "Dewpoint:\t\t" unless $hdr2;
    my @ret = _mkTempVal $t,
                        '', "$hdr1%sC\t\t\t\t\t%sF\n",
                        '', "$hdr2%sC\t\t\t\t\t%sF\n",
                        "relative humidity:\t%.0f%%\n";
    return $ret[0] . $ret[1] . $ret[2];
}

sub _mkRwyVisVal {
    my ($v, $v_gt, $v_lt) = @_;
    my $visVal;

    $visVal = $v->{distance};
    $visVal *= $FT2M if $v->{unitLength} eq 'FT';
    return (undef,
            exists $v->{isEqualGreater} ? $v_gt
                                        : (exists $v->{isLess} ? $v_lt : ''),
            rnd($visVal, 10),
            rnd($visVal / 1000, 0.1),
            rnd($visVal / $STATUTE_MILE2METER, 0.1));
}

sub _printVisRwy {
    my $vis = shift;
    my (@ret, $msg);

    return ('(visibility not available)') if exists $vis->{notAvailable};

    @ret = _mkRwyVisVal $vis, '>=', '<';

    $msg = $ret[1];
    if ($ret[2] >= 1000) {
        $msg .= sprintf "%g km", $ret[3];
    } else {
        $msg .= sprintf "%.0f m", $ret[2];
    }
    $msg .= "\t\t\t\t\t" . $ret[1];
    $msg .= sprintf "%g US-miles\n", $ret[4];
    return $msg;
}

sub _printRwyState {
    my $r = shift;
    my ($msg);

    $msg = '';
    if (exists $r->{SNOCLO}) {
        $msg .= "aerodrome closed for snow\n";
    } else {
        my $delim;
        if (exists $r->{rwyDesigAll}) {
            $msg .= "state all runways";
        } elsif (exists $r->{rwyDesigRep}) {
            $msg .= "state runway REP";
        } else {
            $msg .= 'state runway ' . $r->{rwyDesig};
        }
        $msg .= ":\t";
        $msg .= '(not available)'
            if    exists $r->{depositType}
               && exists $r->{depositExtent}
               && exists $r->{depositDepth}
               && exists $r->{friction}
               && exists $r->{depositType}{notAvailable}
               && exists $r->{depositExtent}{notAvailable}
               && exists $r->{depositDepth}{notAvailable}
               && exists $r->{friction}{notAvailable};
        $delim = '';
        if (exists $r->{cleared}) {
            $msg .= 'cleared';
            $delim = ', ';
        }
        if (   exists $r->{depositType}
            && exists $r->{depositType}{depositTypeVal})
        {
            $msg .= $delim
                    . { 0 => 'clear and dry',
                        1 => 'damp',
                        2 => 'wet or puddles',
                        3 => 'frost',
                        4 => 'dry snow',
                        5 => 'wet snow',
                        6 => 'slush',
                        7 => 'ice',
                        8 => 'compacted snow',
                        9 => 'frozen ridges',
                      }->{$r->{depositType}{depositTypeVal}};
            $delim = ', ';
        }
        if (   exists $r->{depositExtent}
            && exists $r->{depositExtent}{depositExtentVal})
        {
            $msg .= $delim
                    . { 0 => '0%',
                        1 => '1-10%',
                        2 => '11-25%',
                        5 => '26-50%',
                        9 => '51-100%',
                      }->{$r->{depositExtent}{depositExtentVal}}
                    . ($strict_fgfs ? '' : ' of runway covered');
            $delim = ', ';
        }
        if (   exists $r->{depositDepth}
            && exists $r->{depositDepth}{depositDepthVal})
        {
            my $dh_str = $strict_fgfs ? '' : 'deposit depth ';
            $msg .= $delim;
            if ($r->{depositDepth}{depositDepthVal} == 0) {
                $msg .= $dh_str . '<1 mm';
            } elsif ($r->{depositDepth}{depositDepthVal} < 91) {
                $msg .= $dh_str
                        . ($r->{depositDepth}{depositDepthVal} + 0) . ' mm';
            } elsif ($r->{depositDepth} < 99) {
                $msg .= $dh_str
                      . 50 * ($r->{depositDepth}{depositDepthVal} - 90) . ' mm';
            } else {
                $msg .= 'runway not in use';
            }
            $delim = ', ';
        }
        if (exists $r->{friction}) {
            if (exists $r->{friction}{invalidFormat}) {
                $msg .= $delim . "(invalid format: "
                               . "'$r->{friction}{invalidFormat}')";
            } elsif (exists $r->{friction}{coefficient}) {
                $msg .= $delim . 'friction: '
                        . sprintf "%.2f", $r->{friction}{coefficient} / 100;
            } elsif (exists $r->{friction}{brakingAction}) {
                $msg .=  $delim
                         . { 91 => 'poor braking action',
                             92 => 'poor/medium braking action',
                             93 => 'medium braking action',
                             94 => 'medium/good braking action',
                             95 => 'good braking action',
                           } ->{$r->{friction}{brakingAction}};
            } elsif (exists $r->{friction}{unreliable}) {
                $msg .=  $delim . 'friction: unreliable measurement';
            }
        }
        $msg .= "\n";
    }
    return $msg;
}

sub _printColourCode {
    my $c = shift;
    my ($msg, $hdr);
    my $cc = { BLUplus    => 'ceiling 20000 ft, visibility 8 km',
               BLU        => 'ceiling 2500 ft, visibility 8 km',
               WHT        => 'ceiling 1500 ft, visibility 5 km',
               GRN        => 'ceiling 700 ft, visibility 3700 m',
               YLO        => 'ceiling 300 ft, visibility 1600 m',
               YLO1       => 'ceiling ? ft, visibility ? m', # TODO
               YLO2       => 'ceiling ? ft, visibility ? m',
               AMB        => 'ceiling 200 ft, visibility 800 m',
               RED        => 'ceiling <200 ft, visibility <800 m',
               FCSTCANCEL => 'forecast cancelled',
    };

    $hdr .= "Colour code:\t\t";
    if (exists $c->{BLACK}) {
        $msg .= $hdr . "airport closed for technical reasons\n";
        $hdr = "\t\t\t";
    }
    $msg .= $hdr . $cc->{$c->{currentColour}}
        if exists $c->{currentColour};
    if (exists $c->{predictedColour}) {
        $msg .= "\n\t\t\tpredicted:\n\t\t\t";
        if (   exists $c->{currentColour}
            && $c->{currentColour} eq $c->{predictedColour})
        {
            $msg .= 'no change of ceiling or visibility';
        } else {
            $msg .= $cc->{$c->{predictedColour}};
        }
    }

    return $msg;
}

sub _printCloudOpacityLvl {
    my $c = shift;
    my $msg;

    $msg = "phenomenon w. opacity:\t" . $c->{eights} . '/8 ';
    $msg .= $cloud_types{$c->{cloudType}}
        if exists $c->{cloudType};
    $msg .= printWeather($c->{weather}, 1)
        if exists $c->{weather};
    $msg .= ' at ' . $c->{cloudBase} * 100 . ' ft';
    $msg .= printLocations $c->{locationAndList}
        if exists $c->{locationAndList};
    $msg .= "\t\t" . rnd($c->{cloudBase} * 100 * $FT2M, 10) . " m\n";
    return $msg;
}

sub _printUSTemp {
    my ($hdr, $t) = @_;

    return sprintf "$hdr:\t%sC\t\t\t\t\t%sF\n",
                    $t->{temp}, rnd($t->{temp} * 1.8 + 32, 0.1)
        if $t->{tempUnit} eq 'C';
    return sprintf "$hdr:\t%.1fC\t\t\t\t\t%sF\n",
                    rnd(($t->{temp} - 32) / 1.8, 0.1), $t->{temp};
}

sub _mkQNHVal {
    my $qnh = shift;
    my ($qnhinHg);

    return () unless    exists $qnh->{hPa}
                     || exists $qnh->{inHg}
                     || exists $qnh->{mmHg};

    return (0, $qnh->{hPa}, rnd($qnh->{hPa} / $INHG2HPA, 0.01))
        if exists $qnh->{hPa};
    return (0, rnd($qnh->{inHg} * $INHG2HPA, 1), $qnh->{inHg})
        if exists $qnh->{inHg};
    $qnhinHg = $qnh->{mmHg} / ($FT2M / 12 * 1000);
    return (0, rnd($qnhinHg * $INHG2HPA, 1), rnd($qnhinHg, 0.01));
}


sub _printPressure {
    my ($qnh, $hdr, $fmt) = @_;
    my @ret;

    return $hdr . "(not available)\n" if exists $qnh->{notAvailable};
    return "$hdr(invalid format: '$qnh->{invalidFormat}')\n"
        if exists $qnh->{invalidFormat};
    @ret = _mkQNHVal $qnh;
    return $hdr . sprintf $fmt, $ret[1], $ret[2];
}

sub _printTime {
    my ($time) = @_;

    $time =~ '(..)?(..)';
    return ($2 + 0) . " minutes after the hour" unless defined $1;
    return "$1:$2 UTC";
}

sub printMetar_FGFS {
    my ($metar, $opt_m, $strict) = @_;
    my $msg = '';
    my (@ret, $ii, $td, $need_nl, $hdr, $delim);

    $strict_fgfs = $strict;

    $msg .= "msg: $metar->{msg}\n" x $opt_m if defined $opt_m;
    $msg .= $metar->{WARNING} if exists $metar->{WARNING} && !$strict_fgfs;
    if (exists $metar->{SPECI}) {
        $msg .= "SPECI Report";
    } else {
        $msg .= "METAR Report";
    }
    if (exists $metar->{reportModifier}) {
        my $rm = $metar->{reportModifier};
        $msg .= "\t\t(" . { NIL  => 'report missing',
                            AUTO => 'automatically generated',
                            COR  => 'manually corrected',
                            RTD  => 'routine delayed',
                            P    => 'segmented, ',
                            RR   => 'delayed',
                            CC   => 'corrected',
                            AA   => 'amended',
                          }->{$rm->{modifierType}};
        if ($rm->{modifierType} eq 'P') {
            $msg .= 'last ' if exists $rm->{isLastSegment};
            $msg .= "segment $rm->{segment}";
        }
        $msg .= ', bulletin over 24 hours after observation'
            if exists $rm->{over24hLate};
        $msg .= ', bulletin sequence lost'
            if exists $rm->{sequenceLost};
        $msg .= ", bulletin sequence $rm->{bulletinSeq}"
            if exists $rm->{bulletinSeq};
        $msg .= ')';
    }
    $msg .= "\n============\n";
    $msg .= "Airport-Id:\t\t" . $metar->{obsStation}{id} . "\n"
        if exists $metar->{obsStation};

    if (exists $metar->{obsTime} || exists $metar->{issueTime}) {
        my $rep_time;

        if (exists $metar->{obsTime}) {
            $msg .= "Report time:\t\t";
            $rep_time = $metar->{obsTime};
        } else {
            $msg .= "Issue time:\t\t";
            $rep_time = $metar->{issueTime};
        }
        if (exists $rep_time->{invalidFormat}) {
            $msg .= "(invalid format: '$rep_time->{invalidFormat}')\n";
        } else {
            $msg .= sprintf "on the %d., %02d:%02d UTC\n", @$rep_time{'day', 'hour', 'minute'};
        }
    }

    $msg .= "Wind:\t\t\t" . _printWind_FGFS $metar->{sfcWind}{wind}
        if exists $metar->{sfcWind};

    $msg .= $keywords{CAVOK} . "\n"
        if !$strict_fgfs && exists $metar->{CAVOK};

    $msg .= "Visibility:\t\t"
            . _printVis_FGFS('(not available)', $metar->{visPrev}, 1000) . "\n"
        if exists $metar->{visPrev};
    $msg .= "\t\t\t" . _printVis_FGFS(0, $metar->{visMin}, 1000) . "\n"
        if exists $metar->{visMin};

    for (exists $metar->{visRwy} ? @{$metar->{visRwy}} : ()) {
        $need_nl = 0;
        $msg .= "visibility runway " . $_->{rwyDesig} . ":\t";
        if (exists $_->{notAvailable}) {
            $msg .= "(not available)\n";
        } else {
            $msg .= _printVisRwy($_->{RVR});
            if (exists $_->{RVRVariations}) {
                $need_nl = 1;
                $msg .= "\t\t\t" . _printVisRwy($_->{RVRVariations});
            }
        }
        if (!$strict_fgfs && exists $_->{visTrend}) {
            $need_nl = 1;
            $msg .= "\ttrend:\t\t" . $keywords{$_->{visTrend}} . "\n";
        }
        $msg .= "\n" if $need_nl;
    }
    $msg .= $keywords{RVRNO} . "\n"
        if !$strict_fgfs && exists $metar->{RVRNO};

    if (exists $metar->{weather}) {
        $delim = "Weather:\t\t";
        for (@{$metar->{weather}}) {
            $msg .= $delim . printWeather $_;
            $delim = ', ';
        }
        $msg .= "\n";
    }

    $msg .= _printCloud $metar if exists $metar->{cloud};

    $msg .= "Vert. visibility:\t"
            . _printVis_FGFS('impossible to determine', $metar->{visVert}, 1000)
            . "\n"
        if exists $metar->{visVert};

    $msg .= _printTemp_FGFS($metar->{temperature})
        if exists $metar->{temperature};

    if (exists $metar->{QNH}) {
        @ret = _mkQNHVal $metar->{QNH};
        $msg .= sprintf "Pressure:\t\t%.0f hPa\t\t\t\t%g in. Hg\n",
                        $ret[1], $ret[2]
            if $ret[1];
    }

    if (!$strict_fgfs && exists $metar->{somePressure}) {
        @ret = _mkQNHVal $metar->{somePressure};
        $msg .= sprintf "Some pressure:\t\t%.0f hPa\t\t\t\t%g in. Hg\n",
                        $ret[1], $ret[2]
            if $ret[1];
    }

    for (exists $metar->{windShear} ? @{$metar->{windShear}} : ()) {
        $msg .=   (exists $_->{rwyDesigAll}
                    ? 'all runways' : "runway $_->{rwyDesig}")
                . ":\t\tcritical wind shear\n";
    }

    for (exists $metar->{rwyState} ? @{$metar->{rwyState}} : ()) {
        $msg .= _printRwyState $_;
    }

    if (   $strict_fgfs
        && exists $metar->{trend}
        && $metar->{trend}[0]{trendType} eq 'NOSIG')
    {
        for (exists $metar->{trend}[0]{rwyState}
             ? @{$metar->{trend}[0]{rwyState}} : ())
        {
            $msg .= _printRwyState $_;
        }
    }

    if (!$strict_fgfs && exists $metar->{recWeather}) {
        $delim = "recent weather:\t\t";
        for (@{$metar->{recWeather}}) {
            $msg .= $delim . printWeather($_, 1);
            $delim = ', ';
        }
        $msg .= "\n";
    }

    $msg .= _printColourCode($metar->{colourCode}) . "\n"
        if !$strict_fgfs && $metar->{colourCode};

    $msg .= "NEFO PLAYA:\t\t" . $metar->{NEFO_PLAYA} . "\n"
        if exists $metar->{NEFO_PLAYA} && !$strict_fgfs;

    for (exists $metar->{rwyWind} && !$strict_fgfs ? @{$metar->{rwyWind}} : ()){
        $msg .= "wind for runway $_->{rwyDesig}:\t"
                . _printWind_FGFS $_->{wind};
    }

    $hdr = "trends within the next 2 hours:\n===============================\n";
    for my $td (exists $metar->{trend} && !$strict_fgfs ? @{$metar->{trend}} : ())
    {
        $msg .= "\n$hdr";
        $hdr = '';
        if (exists $td->{trendTime1}) {
            $msg .= $keywords{$td->{trendTime1}{timeSpec}} . ' '
                    . $td->{trendTime1}{hour} . ':'
                    . $td->{trendTime1}{minute} . ' UTC ';
            $msg .= $keywords{$td->{trendTime2}{timeSpec}} . ' '
                    . $td->{trendTime2}{hour} . ':'
                    . $td->{trendTime2}{minute} . ' UTC '
                if exists $td->{trendTime2};
        }
        $msg .=  $keywords{$td->{trendType}} . "\n";

        $msg .= $keywords{CAVOK} . "\n"
            if exists $td->{CAVOK};

        $msg .= "Visibility:\t\t" . _printVis_FGFS('(not available)',
                                                  $td->{visPrev}, 1000) . "\n"
            if exists $td->{visPrev};

        $msg .= "Vert. visibility:\t"
               . _printVis_FGFS('impossible to determine', $td->{visVert}, 1000)
               . "\n"
            if exists $td->{visVert};

        $msg .= "Wind:\t\t\t" . _printWind_FGFS $td->{sfcWind}{wind}
            if exists $td->{sfcWind};

        if (exists $td->{weather}) {
            $delim = "Weather:\t\t";
            for (@{$td->{weather}}) {
                $msg .= $delim . printWeather $_;
                $delim = ', ';
            }
            $msg .= "\n";
        }

        $msg .= _printCloud $td if exists $td->{cloud};

        for (exists $td->{rwyState} ? @{$td->{rwyState}} : ()) {
            $msg .= _printRwyState $_;
        }

        $msg .= _printColourCode($td->{colourCode}) . "\n"
            if $td->{colourCode};
    }

    if (!$strict_fgfs && exists $metar->{remark}) {
        my $rem = $metar->{remark};
        $msg .= "\nRemarks:\n========\n";

        for (exists $metar->{remark} ? @{$metar->{remark}} : '') {
            my $r = $_;
            if (exists $r->{freetext}) {
                $msg .= $r->{freetext}{txt} . "\n";
                next;
            }
            for (sort keys %$r) {
                my $e = $r->{$_};
                if ($_ eq 'obsStationType') {
                    $msg .= "automated station "
                             . { AO1 => 'without',
                                 AO2 => 'with' }->{$e->{stationType}}
                             . " precipitation discriminator\n";
                } elsif ($_ eq 'visAt') {
                    if (exists $e->{locationAt}) {
                        $msg .= "Visibility to the sea:\t"
                            if $e->{locationAt} eq 'MAR';
                        $msg .= "visibility at "
                                . $keywords{$e->{locationAt}} . ":\t"
                            if $e->{locationAt} =~ 'SFC|TWR';
                    } else {
                        $msg .= "visibility:\t";
                    }
                    $msg .= _printVis_FGFS(0, $e->{visibility}, 1000) ."\n";
                } elsif ($_ eq 'visVar1') {
                    $msg .= "visibility varies from:\t"
                            . _printVis_FGFS(0, $e, 1000) . "\n";
                } elsif ($_ eq 'visVar2') {
                    $msg .= "\t\tto:\t" . _printVis_FGFS(0, $e, 1000) . "\n";
                } elsif ($_ eq 'cloudMaxCover') {
                    $msg .= "max. cloud cover:\t"
                            . _printCloudCover($e->{cloudCover}) . "\n";
                } elsif ($_ eq 'cloudOpacityLvl') {
                    $msg .= _printCloudOpacityLvl $e;
                } elsif ($_ eq 'rwyState') {
                    $msg .= _printRwyState $e;
                } elsif ($_ eq 'recWeather') {
                    $msg .= "recent weather:\t\t"
                            . printWeather($e->[0], 1) . "\n";
                } elsif ($_ eq 'visMin') {
                    $msg .= "minimum visibility:\t"
                            . _printVis_FGFS(0, $e, 1000) . "\n";
                } elsif ($_ eq 'visListLoc') {
                    $delim = $keywords{VIS};
                    for (@{$e->{arr}}) {
                        $msg .= $delim;
                        $delim = "\n" . ' ' x 10;
                        $msg .= printLocations($_->{locationAndList}) . ":\t"
                              . _printVis_FGFS(0, $_->{visibility}, 1000);
                    }
                    $msg .= "\n";
                } elsif ($_ eq 'correctedAt') {
                    $msg .= "report corrected at:\t$e->{hour}:$e->{minute} UTC\n";
                } elsif ($_ eq 'colourCode') {
                    $msg .= _printColourCode($e) . "\n";
                } elsif ($_ eq 'QNH') {
                    $msg .= _printPressure $e,
                                     "Pressure:\t\t",
                                     "%.0f hPa\t\t\t\t%g in. Hg\n";
                } elsif ($_ =~ '^SLP') {
                    $msg .= _printPressure $e,
                                     $keywords{SLP} . ":\t",
                                     "%.1f hPa\t\t\t\t%g in. Hg\n";
                } elsif ($_ eq 'regQNH') {
                    $msg .= _printPressure $e,
                                          "reg. QNH:\t\t",
                                          "%.0f hPa\t\t\t\t%g in. Hg\n";
                } elsif ($_ eq 'QFE') {
                    $msg .= _printPressure $e,
                                          "QFE pressure:\t\t",
                                          "%.0f hPa\t\t\t\t%g in. Hg\n";
                } elsif ($_ eq 'QFF') {
                    $msg .= _printPressure $e,
                                          "QFF pressure:\t\t",
                                          "%.0f hPa\t\t\t\t%g in. Hg\n";
                } elsif ($_ eq 'reportConcerns') {
                    $msg .= 'report concerns '
                           . { M => 'deterioration',
                               B => 'improvement'
                             }->{$e->{change}}
                           . ' of '
                           . { 0 => 'gust',
                               1 => $keywords{WND},
                               2 => $keywords{VIS},
                               3 => $keywords{CLD},
                               4 => $keywords{PCPN},
                               5 => 'pressure',
                               6 => 'state of sea or swell',
                               7 => $weather_types{DS} . ', ' .
                                    $weather_types{SS} . ' or ' .
                                    $weather_descr{BL} . ' ' .
                                    $weather_types{SN},
                               8 => $weather_types{TS},
                               9 => $weather_types{SQ} . ' or tornado',
                             }->{$e->{subject}};
                } elsif ($_ eq 'RH') {
                    $msg .= "$keywords{RH}:\t" . $e->{RHVal} . "%\n";
                } elsif ($_ eq 'AI') {
                    $msg .= "$keywords{AI}:\t\t\t$e->{AIVal}\n";
                } elsif ($_ eq 'SST' || $_ eq 'OAT') {
                    $msg .= _printUSTemp $keywords{$_}, $e;
                } elsif ($_ eq 'temp6hMax') {
                    $msg .= _printUSTemp '6h max. temperature', $e;
                } elsif ($_ eq 'temp6hMin') {
                    $msg .= _printUSTemp '6h min. temperature', $e;
                } elsif ($_ eq 'temp24h') {
                    $msg .= _printUSTemp '24h max. temperature',
                                         $e->{temp24hMax};
                    $msg .= _printUSTemp '24h min. temperature',
                                         $e->{temp24hMin}
                        if exists $e->{temp24hMin};
                } elsif ($_ eq 'tempHourly') {
                    $msg .= _printTemp_FGFS $e, "1h avg. temperature:\t",
                                                "1h avg. dewpoint:\t";
                } elsif ($_ eq 'peakWind') {
                    $msg .= 'peak wind at '
                            .   _printTime((exists $e->{hour} ? $e->{hour} : '')
                                          . $e->{minute})
                            . ":\t";
                    $msg .= _printWind_FGFS $e->{wind};
                } elsif ($_ eq 'visRwy') {
                    $msg .= "visibility runway " . $e->{rwyDesig} . ":\t"
                            . _printVisRwy($e->{RVR});
                } elsif ($_ eq 'rwyWind') {
                    $msg .= "wind for runway $e->{rwyDesig}:\t"
                            . _printWind_FGFS $e->{wind};
                } elsif ($_ eq 'thrWind') {
                    $msg .= "wind for runway $e->{rwyDesig}:\t"
                            . _printWind_FGFS $e->{wind};
                } elsif ($_ eq 'cloudCoverVar') {
                    $msg .= "Sky condition:\t\t"
                            . _printCloudCover($e->{cloudCover}) . "\n"
                            . "varying to:\t\t\t"
                            . _printCloudCover($e->{cloudCover2}) . "\n"
                } elsif ($_ eq 'precipHourly') {
                    if (exists $e->{precipHoursNotAvailable}) {
                        $msg .= '3-/6-hour precip. amount';
                    } elsif ($e->{precipHours} == 1) {
                        $msg .= 'hourly precip. amount';
                    } else {
                        $msg .= $e->{precipHours} . '-hour precip. amount';
                    }
                    $msg .= ":\t";
                    if (exists $e->{notAvailable}) {
                        $msg .= "(not available)\n";
                    } else {
                        $msg .= sprintf "%.0f mm\t\t\t\t\t%g in.\n",
                                $e->{precipVal} * ($FT2M / 12 * 1000),
                                $e->{precipVal};
                    }
                } elsif ($_ eq 'snowIncr') {
                    $msg .= "snow increasing rapidly:\t" . $e->{pastHour}
                            . " in. past hour, " . $e->{onGround}
                            . " in. on ground\n";
                } elsif ($_ =~ '^(PWINO|FZRANO|TSNO|PNO|RVRNO|PRES[FR]R|AURBO|NOSPECI|RCRNR)$')
                {
                    $msg .= $keywords{$1} . "\n";
                } elsif ($_ eq 'needMaint') {
                    $msg .= $keywords{'$'} . "\n";
                } elsif ($_ eq 'cloud') {
                    $msg .= _printCloud { cloud => [$e] };
                } elsif ($_ eq 'cloudTypeFamily') {
                    $msg .= "cloud types:\n\tlow:\t\t";
                    if (exists $e->{cloudTypeLowAboveOvercast}) {
                        $msg .= "layer above overcast\n";
                    } else {
                        $msg .= $cloud_types_low{$e->{cloudTypeLow}} . "\n";
                    }
                    $msg .= "\tmid-level:\t";
                    if (exists $e->{cloudTypeMiddleAboveOvercast}) {
                        $msg .= "layer above overcast\n";
                    } else {
                        $msg .= $cloud_types_middle{$e->{cloudTypeMiddle}}."\n";
                    }
                    $msg .= "\thigh:\t\t";
                    if (exists $e->{cloudTypeHighAboveOvercast}) {
                        $msg .= "layer above overcast\n";
                    } else {
                        $msg .= $cloud_types_high{$e->{cloudTypeHigh}} . "\n";
                    }
                } elsif ($_ eq 'windShift') {
                    $msg .= "wind shift at "
                            .   _printTime((exists $e->{hour} ? $e->{hour} : '')
                                          . $e->{minute});
                    $msg .= ' due to frontal passage' if exists $e->{FROPA};
                    $msg .= "\n";
                } elsif ($_ eq 'beginEndPrecip') {
                    my ($txt_loc_mov, $hdr);

                    if (   exists $e->{locationAndList}
                        || exists $e->{MOV}
                        || exists $e->{MOVD})
                    {
                        $txt_loc_mov = "\n\t\t\t";
                        if (exists $e->{locationAndList}) {
                            $txt_loc_mov .=
                                          printLocations($e->{locationAndList});
                            $txt_loc_mov .= ','
                                if exists $e->{MOV} || exists $e->{MOVD};
                        }
                        $txt_loc_mov .= " $keywords{MOV}"
                                        . printLocations $e->{MOV}
                            if exists $e->{MOV};
                        $txt_loc_mov .= " $keywords{MOVD}"
                                        . printLocations $e->{MOVD}
                            if exists $e->{MOVD};
                    }

                    $hdr = "precipitation:\t";
                    for (@{$e->{precip}}) {
                        $msg .= $hdr . printWeather($_->{weather}, 1) . ':';
                        for (@{$_->{start_end}}) {
                            $msg .= "\n\t\t\t\t";
                            if (exists $_->{beginPrecip}) {
                                $msg .= 'began at '
                                  .   _printTime((exists $_->{beginPrecip}{hour}
                                                 ? $_->{beginPrecip}{hour} : '')
                                                . $_->{beginPrecip}{minute});
                            } else {
                                $msg .= 'ended at '
                                  .   _printTime((exists $_->{endPrecip}{hour}
                                                 ? $_->{endPrecip}{hour} : '')
                                                . $_->{endPrecip}{minute});
                            }
                        }
                        $hdr = "\n\t\t\t";
                    }
                    $msg .= $txt_loc_mov if defined $txt_loc_mov;
                    $msg .= "\n";
                } elsif ($_ eq 'visLower') {
                    $msg .= "@keywords{'VIS', 'LWR'}"
                            . printLocations($e->{locationAndList}) . "\n";
                } elsif ($_ eq 'cigLower') {
                    $msg .= "@keywords{'CIG', 'LWR'}"
                            . printLocations($e->{locationAndList}) . "\n";
                } elsif ($_ eq 'phenomLocation') {
                    $msg .= "$keywords{LOW} " if exists $e->{isLow};
                    $msg .= "$keywords{OCNL} " if exists $e->{isOccasional};
                    $msg .= "$keywords{LGT} " if exists $e->{isLight};
                    $msg .= "@keywords{'VRY', 'LGT'} "
                        if exists $e->{isVeryLight};
                    $msg .= "$keywords{INTMT} " if exists $e->{isIntermittent};
                    $msg .= "$keywords{CONS} " if exists $e->{isContinuous};
                    $hdr = '';
                    for (exists $e->{cloudType} ? @{$e->{cloudType}} : ()) {
                        $msg .= $hdr . $cloud_types{$_};
                        $hdr .= ' and ';
                    }
                    $msg .= printWeather $e->{weather}, 1
                        if exists $e->{weather};
                    $msg .= _printCloudCover $e->{cloudCover}
                        if exists $e->{cloudCover};
                    $msg .= $keywords{$e->{otherPhenom}}
                        if exists $e->{otherPhenom};
                    $msg .= " $keywords{'PAST HR'}"
                        if exists $e->{inPastHour};
                    $msg .= " from $keywords{CONTRAILS}"
                        if exists $e->{fromContrails};
                    if (exists $e->{locationAndList}) {
                        $msg .= printLocations($e->{locationAndList});
                        $msg .= ',' if exists $e->{MOV} || exists $e->{MOVD};
                    }
                    $msg .= " $keywords{MOV}" . printLocations $e->{MOV}
                        if exists $e->{MOV};
                    $msg .= " $keywords{MOVD}" . printLocations $e->{MOVD}
                        if exists $e->{MOVD};
                    $msg .= "\n";
                } elsif ($_ eq 'conditionMountain') {
                    my %specificatione_mon = (
                        LIB       => 'free of clouds',
                       'CLD SCT'  => 'partially covered in scattered clouds',
                       'VERS_INC' => 'slopes in clouds, top is free',
                       'CNS_POST' => 'free from side of observer, other side in clouds',
                       'CLD_CIME' => 'clouds touch top',
                       'CIME_INC' => 'top in clouds, slopes free',
                       'GEN_INC'  => 'generally in clouds, some tops free',
                        INC       => 'in clouds',
                        INVIS     => 'not visible',
                    );
                    my %evoluzione_mon = (
                        NC          => 'no change',
                        CUF         => 'cumuli developing',
                       'ELEV_SLW'   => 'developing slowly',
                       'ELEV_RAPID' => 'developing rapidly',
                       'ELEV_STF'   => 'developing and stratification',
                       'ABB_SLW'    => 'sinking slowly',
                       'ABB_RAPID'  => 'sinking rapidly',
                        STF         => 'stratification',
                       'STF_ABB'    => 'stratification and sinking',
                       'VAR_RAPID'  => 'varying rapidly',
                    );

                    $msg .= "$keywords{MTNS}:\t\t";
                    $delim = '';
                    for (@{$e->{condMoun}}) {
                        $msg .= $delim;
                        $delim = ', ';
                        $msg .= printLocations($_->{locationAndList}) . ': '
                            if defined $_->{locationAndList};
                        $msg .= $specificatione_mon{$_->{condMounType}};
                        $msg .= ', ' . $evoluzione_mon{$_->{condMounChange}}
                            if defined $_->{condMounChange};
                    }
                    $msg .= "\n";
                } elsif ($_ eq 'conditionValley') {
                    my %specificatione_val = (
                        NIL                 => 'no low clouds, fog or haze',
                        FOSCHIA             => 'haze',
                       'FOSCHIA_SKC_SUP'    => 'haze, sky clear above',
                        NEBBIA              => 'fog',
                       'NEBBIA_SCT'         => 'scattered fog banks',
                       'CLD_SCT'            => 'isolated clouds',
                       'CLD_SCT_NEBBIA_INF' => 'isolated clouds, fog below',
                       'MAR_CLD'            => 'sea in clouds',
                        INVIS               => 'not visible',
                    );
                    my %evoluzione_val = (
                        NC            => 'no change',
                        DIM           => 'decreasing',
                       'DIM_ELEV'     => 'decreasing and raising',
                       'DIM_ABB'      => 'decreasing and sinking',
                        AUM           => 'increasing',
                       'AUM_ELEV'     => 'increasing and raising',
                       'AUM_ABB'      => 'increasing and sinking',
                        ELEV          => 'raising',
                        ABB           => 'sinking',
                       'NEBBIA_INTER' => 'intermittent fog at station',
                    );

                    $msg .= "$keywords{VAL}:\t\t";
                    $delim = '';
                    for (@{$e->{condVall}}) {
                        $msg .= $delim;
                        $delim = ', ';
                        $msg .= printLocations($_->{locationAndList}) . ': '
                            if defined $_->{locationAndList};
                        $msg .= $specificatione_val{$_->{condVallType}};
                        $msg .= ', ' . $evoluzione_val{$_->{condVallChange}}
                            if defined $_->{condVallChange};
                    }
                    $msg .= "\n";
                }
            }
        }
    }

    $msg .= "ERROR: " . { obsStation => 'not a station code',
                        obsTime    => 'not an observation time',
                        other      => 'invalid format',
                      }->{$metar->{ERROR}{descr}}
                    . ': ' . $metar->{ERROR}{pos} . "\n"
        if exists $metar->{ERROR};

    return $msg;
}

1;
__END__
