package metaf2xml::XML;

# copyright (c) metaf2xml 2006
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

use strict;
use warnings;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
require Exporter;

@ISA = qw(Exporter);

@EXPORT = qw(
    xml_start
    printReport_XML
    xml_end
);

sub _mkXMLTag {
    my ($root, $attr_name) = @_;
    my @attr;

    for (sort keys %{$root->{$attr_name}}) {
        push @attr, ($_,  $root->{$attr_name}{$_});
    }
    return @attr;
}

sub _printXMLTag {
    my ($metar, $writer, $tag, $skip_s) = @_;
    $writer->startTag($tag, !$skip_s && exists $metar->{$tag}{s}
                            ? ('s', $metar->{$tag}{s}) : ());
    for (sort keys %{$metar->{$tag}}) {
        next if $_ eq 's';
        if ($tag eq 'rwyState' && ($_ eq 'friction' || $_ =~ /^deposit/)) {
            _printXMLTag($metar->{$tag}, $writer, $_);
        } else {
            $writer->emptyTag($_, 'v', $metar->{$tag}{$_});
        }
    }
    $writer->endTag($tag);
}

sub _printXMLTagList {
    my ($metar, $writer, $tag) = @_;
    $writer->startTag($tag . 'List');
    for (@{$metar->{$tag}}) {
        _printXMLTag({ $tag => $_ }, $writer, $tag);
    }
    $writer->endTag($tag . 'List');
}

sub _printXMLWeather {
    my ($weather, $writer, $tag) = @_;

    $writer->startTag($tag) if defined $tag;
    for (@$weather) {
        $writer->startTag('weather', 's', $_->{s});
        $writer->emptyTag('notAvailable', 'v', 1) if exists $_->{notAvailable};
        $writer->emptyTag('invalidFormat', 'v', $_->{invalidFormat})
            if exists $_->{invalidFormat};
        $writer->emptyTag('NSW', 'v', 1) if exists $_->{NSW};
        $writer->emptyTag('inVicinity', 'v', 1) if exists $_->{inVicinity};
        $writer->emptyTag('intensity', 'v', $_->{intensity})
            if exists $_->{intensity};
        for (exists $_->{descriptor} ? @{$_->{descriptor}} : ()) {
            $writer->emptyTag('descriptor', 'v', $_);
        }
        for (exists $_->{phenomenon} ? @{$_->{phenomenon}} : ()) {
            $writer->emptyTag('phenomenon', 'v', $_);
        }
        $writer->endTag('weather');
    }
    $writer->endTag($tag) if defined $tag;
}

sub _printXMLWind {
    my ($wind, $writer, $tag) = @_;

    $writer->startTag($tag, exists $wind->{s} ? ('s', $wind->{s}) : ());
    for (sort keys %$wind) {
        next if $_ eq 's';
        if ($_ eq 'wind') {
            _printXMLTag($wind, $writer, $_);
        } else {
            $writer->emptyTag($_, 'v', $wind->{$_});
        }
    }
    $writer->endTag($tag);
}

sub _printXMLLocations {
    my ($metar, $writer) = @_;

    $writer->startTag('locationAndList');
    for (@$metar) {
        $writer->startTag('locationThruList');
        for (@$_) {
            $writer->startTag('location');
            $writer->emptyTag('isDistant', 'v', 1)  if exists $_->{isDistant};
            $writer->emptyTag('inVicinity', 'v', 1) if exists $_->{inVicinity};
            if (exists $_->{distance}) {
                $writer->emptyTag('distance', 'v', $_->{distance});
                $writer->emptyTag('unitLength', 'v', $_->{unitLength});
            }
            $writer->emptyTag('isGrid', 'v', 1)     if exists $_->{isGrid};
            $writer->emptyTag('isQuadrant', 'v', 1) if exists $_->{isQuadrant};
            $writer->emptyTag('obscgMtns', 'v', 1)  if exists $_->{obscgMtns};
            if (exists $_->{locationSpec}) {
                $writer->emptyTag('locationSpec', 'v', $_->{locationSpec});
            } elsif (exists $_->{compassDir}) {
                $writer->emptyTag('compassDir', 'v', $_->{compassDir});
            } else {
                $writer->startTag('quadrantList');
                for (@{$_->{quadrant}}) {
                    $writer->emptyTag('quadrant', 'v', $_);
                }
                $writer->endTag('quadrantList');
            }
            $writer->endTag('location');
        }
        $writer->endTag('locationThruList');
    }
    $writer->endTag('locationAndList');
}

sub _printXMLLocSpecs {
    my ($metar, $writer) = @_;
    _printXMLLocations($metar->{locationAndList}, $writer)
        if exists $metar->{locationAndList};
    if (exists $metar->{MOV}) {
        $writer->startTag('MOV');
        _printXMLLocations($metar->{MOV}, $writer);
        $writer->endTag('MOV');
    } elsif (exists $metar->{MOVD}) {
        $writer->startTag('MOVD');
        _printXMLLocations($metar->{MOVD}, $writer);
        $writer->endTag('MOVD');
    }
}

sub printReport_XML {
    my ($metar, $writer, $is_taf) = @_;

    $writer->startTag($is_taf ? 'taf' : 'metar', 's', $metar->{msg});
    _printXMLTag($metar, $writer, 'ERROR')
        if exists $metar->{ERROR};
    $writer->dataElement('WARNING', ($metar->{WARNING} =~ /(.*?)\n?$/s))
        if $metar->{WARNING} ne '';
    $writer->emptyTag('SPECI', 'v', 1) if exists $metar->{SPECI};
    $writer->emptyTag('correctedMsg', 'v', $metar->{msgCorr})
        if $metar->{msg} . ' ' ne $metar->{msgCorr};
    _printXMLTag($metar, $writer, 'obsStation')
        if exists $metar->{obsStation};
    _printXMLTag($metar, $writer, 'obsTime')
        if exists $metar->{obsTime};
    _printXMLTag($metar, $writer, 'issueTime')
        if exists $metar->{issueTime};
    _printXMLTag($metar, $writer, 'fcstPeriod')
        if exists $metar->{fcstPeriod};
    _printXMLTag($metar, $writer, 'reportModifier')
        if exists $metar->{reportModifier};
    _printXMLWind($metar->{sfcWind}, $writer, 'sfcWind')
        if exists $metar->{sfcWind};
    $writer->emptyTag('CAVOK') if exists $metar->{CAVOK};
    _printXMLTag($metar, $writer, 'visPrev') if exists $metar->{visPrev};
    _printXMLTag($metar, $writer, 'visMin') if exists $metar->{visMin};
    if (exists $metar->{visRwy}) {
        $writer->startTag('visRwyList');
        for my $v (@{$metar->{visRwy}}) {
            $writer->startTag('visRwy', 's', $v->{s});
            for (sort keys %$v) {
                next if $_ eq 's';
                if ($_ eq 'RVR' || $_ eq 'RVRVariations') {
                    _printXMLTag($v, $writer, $_);
                } else {
                    $writer->emptyTag($_, 'v', $v->{$_});
                }
            }
            $writer->endTag('visRwy');
        }
        $writer->endTag('visRwyList');
    }
    $writer->emptyTag('RVRNO') if exists $metar->{RVRNO};
    _printXMLWeather $metar->{weather}, $writer, 'weatherList'
        if exists $metar->{weather};
    _printXMLTagList($metar, $writer, 'cloud') if exists $metar->{cloud};
    _printXMLTag($metar, $writer, 'visVert') if exists $metar->{visVert};
    if (exists $metar->{temperature}) {
        $writer->startTag('temperature', 's', $metar->{temperature}{s});
        _printXMLTag($metar->{temperature}, $writer, 'air');
        _printXMLTag($metar->{temperature}, $writer, 'dewpoint')
            if exists $metar->{temperature}{dewpoint};
        if (exists $metar->{temperature}{relHumid1}) {
            for (1, 2, 3, 4) {
              $writer->emptyTag('relHumid' . $_,
                                'v', $metar->{temperature}{'relHumid' . $_})
            }
        }
        $writer->endTag('temperature');
    }
    if (exists $metar->{tempMaxMin}) {
        _printXMLTag($metar, $writer, 'tempMax') if exists $metar->{tempMax};
        _printXMLTag($metar, $writer, 'tempMin') if exists $metar->{tempMin};
    } else {
        _printXMLTag($metar, $writer, 'tempMin') if exists $metar->{tempMin};
        _printXMLTag($metar, $writer, 'tempMax') if exists $metar->{tempMax};
    }
    _printXMLTagList($metar, $writer, 'tempAt') if exists $metar->{tempAt};
    _printXMLTag($metar, $writer, 'QNH') if exists $metar->{QNH};
    _printXMLTag($metar, $writer, 'somePressure')
        if exists $metar->{somePressure};
    _printXMLTag($metar, $writer, 'cloudMaxCover')
        if exists $metar->{cloudMaxCover};
    _printXMLWeather $metar->{recWeather}, $writer, 'recentWeatherList'
        if exists $metar->{recWeather};
    _printXMLTagList($metar, $writer, 'windShear')
        if exists $metar->{windShear};
    _printXMLTagList($metar, $writer, 'rwyState')
        if exists $metar->{rwyState};
    _printXMLTag($metar, $writer, 'colourCode')
        if exists $metar->{colourCode};
    if (exists $metar->{rwyWind}) {
        $writer->startTag('rwyWindList');
        for (@{$metar->{rwyWind}}) {
            _printXMLWind($_, $writer, 'rwyWind');
        }
        $writer->endTag('rwyWindList');
    }
    for my $suppl (@{$metar->{TAFsuppl}}) {
        my $r_type = (keys %{$suppl})[0];
        if ($r_type eq 'windShearLvl') {
            $writer->startTag('windShearLvl', 's', $metar->{windShearLvl}{s});
            $writer->emptyTag('level', 'v', $metar->{windShearLvl}{level});
            _printXMLWind($metar->{windShearLvl}{wind}, $writer, 'wind');
            $writer->endTag('windShearLvl');
        } elsif ($r_type =~ /^(?:turbulence|icing)$/) {
            _printXMLTag($suppl, $writer, $r_type);
        }
    }

    if (exists $metar->{trend}) {
        $writer->startTag('trendList');
        for my $td (@{$metar->{trend}}) {
            $writer->startTag('trend', 's', $td->{s});
            $writer->emptyTag('trendType', 'v', $td->{trendType});
            _printXMLTag($td, $writer,'trendTime1') if exists $td->{trendTime1};
            _printXMLTag($td, $writer,'trendTime2') if exists $td->{trendTime2};
            $writer->emptyTag('probability', 'v', $td->{probability})
                if exists $td->{probability};
            _printXMLWind($td->{sfcWind}, $writer, 'sfcWind')
                if exists $td->{sfcWind};
            $writer->emptyTag('CAVOK') if exists $td->{CAVOK};
            _printXMLTag($td, $writer, 'visPrev') if exists $td->{visPrev};
            _printXMLTag($td, $writer, 'visVert') if exists $td->{visVert};
            _printXMLWeather $td->{weather}, $writer, 'weatherList'
                if exists $td->{weather};
            _printXMLTagList($td, $writer, 'cloud') if exists $td->{cloud};
            _printXMLTagList($td, $writer,'rwyState') if exists $td->{rwyState};
            _printXMLTag($td, $writer,'colourCode') if exists $td->{colourCode};
            for my $suppl (@{$td->{TAFsuppl}}) {
                my $r_type = (keys %{$suppl})[0];
                if ($r_type eq 'windShearLvl') {
                    $writer->startTag('windShearLvl', 's',
                                      $metar->{windShearLvl}{s});
                    $writer->emptyTag('level', 'v',
                                      $metar->{windShearLvl}{level});
                    _printXMLWind($metar->{windShearLvl}{wind}, $writer,'wind');
                    $writer->endTag('windShearLvl');
                } elsif ($r_type =~ /^(?:turbulence|icing)$/) {
                    _printXMLTag($suppl, $writer, $r_type);
                }
            }

            $writer->endTag('trend');
        }
        $writer->endTag('trendList');
    }

    if (exists $metar->{remark}) {
        $writer->startTag('remarkList');
        for my $r (@{$metar->{remark}}) {
            $writer->startTag('remark');
            $writer->dataElement('freetext',
                              ($r->{freetext}{txt} =~ /(.*?)\n?$/s),
                              's', $r->{freetext}{s}),
                if exists $r->{freetext};
            for my $r_type (sort keys %$r) {
                if ($r_type =~ '^(?:obsStationType|wetRunway|visVar.|cloudMaxCover|cloudCoverVar|cloud|colourCode|QNH|SLP.*|regQNH|QFE|QFF|windShift|snowIncr|correctedAt|rwyState|reportConcerns|temp6hMax|temp6hMin|precipHourly|AI|RH|OAT|SST|cloudTypeFamily)$') {
                    _printXMLTag($r, $writer, $r_type);
                } elsif ($r_type =~ '^(?:recWeather)$') {
                    _printXMLWeather $r->{$r_type}, $writer,'recentWeatherList';
                } elsif ($r_type =~ '^(?:visAt|visMin)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (sort keys %{$r->{$r_type}}) {
                        next if $_ eq 's';
                        if ($_ eq 'visibility') {
                            _printXMLTag($r->{$r_type}, $writer, $_);
                        } else {
                            $writer->emptyTag($_, 'v', $r->{$r_type}{$_});
                        }
                    }
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:thrWind|peakWind|rwyWind|gridWind)$') {
                    _printXMLWind($r->{$r_type}, $writer, $r_type);
                } elsif ($r_type =~ '^(?:visListLoc)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (@{$r->{$r_type}{arr}}) {
                        $writer->startTag('visLocData');
                        _printXMLTag($_, $writer, 'visibility');
                        _printXMLLocSpecs $_, $writer;
                        $writer->endTag('visLocData');
                    }
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:cloudOpacityLvl)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    _printXMLLocSpecs($r->{$r_type}, $writer);
                    $writer->emptyTag('eights', 'v', $r->{$r_type}{eights});
                    _printXMLWeather [$r->{$r_type}{weather}], $writer
                        if exists $r->{$r_type}{weather};
                    $writer->emptyTag('cloudType', 'v',$r->{$r_type}{cloudType})
                        if exists $r->{$r_type}{cloudType};
                    $writer->emptyTag('cloudBase', 'v',
                                                      $r->{$r_type}{cloudBase});
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:cloudAbove)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (sort keys %{$r->{$r_type}}) {
                        next if $_ eq 's';
                        if ($_ eq 'cloud') {
                            _printXMLTag($r->{$r_type}, $writer, $_);
                        } else {
                            $writer->emptyTag($_, 'v', $r->{$r_type}{$_});
                        }
                    }
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:visRwy)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (sort keys %{$r->{$r_type}}) {
                        next if $_ eq 's';
                        if ($_ eq 'RVR') {
                            _printXMLTag($r->{$r_type}, $writer, $_);
                        } else {
                            $writer->emptyTag($_, 'v', $r->{$r_type}{$_});
                        }
                    }
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:tempHourly)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    _printXMLTag($r->{$r_type}, $writer, 'air');
                    _printXMLTag($r->{$r_type}, $writer, 'dewpoint')
                        if exists $r->{$r_type}{dewpoint};
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:temp24h)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    _printXMLTag($r->{$r_type}, $writer, 'temp24hMax');
                    _printXMLTag($r->{$r_type}, $writer, 'temp24hMin')
                        if exists $r->{$r_type}{temp24hMin};
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:PWINO|FZRANO|TSNO|PNO|RVRNO|PRES[FR]R|AURBO|NOSPECI|RCRNR)$') {
                    $writer->emptyTag($r_type);
                } elsif ($r_type eq 'needMaint') {
                    $writer->emptyTag($r_type, 's', '$');
                } elsif ($r_type eq 'beginEndPrecip') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (@{$r->{$r_type}{precip}}) {
                        $writer->startTag('precip');
                        for (exists $_->{weather}{descriptor}
                                    ? @{$_->{weather}{descriptor}} : ()) {
                            $writer->emptyTag('descriptor', 'v', $_);
                        }
                        for (exists $_->{weather}{phenomenon}
                                    ? @{$_->{weather}{phenomenon}} : ()) {
                            $writer->emptyTag('phenomenon', 'v', $_);
                        }
                        for (@{$_->{start_end}}) {
                            if (exists $_->{beginPrecip}) {
                                _printXMLTag($_, $writer, 'beginPrecip');
                            } else {
                                _printXMLTag($_, $writer, 'endPrecip');
                            }
                        }
                        $writer->endTag('precip');
                    }
                    _printXMLLocSpecs $r->{$r_type}, $writer;
                    $writer->endTag($r_type);
                } elsif ($r_type =~ '^(?:visLower|cigLower)$') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    _printXMLLocSpecs($r->{$r_type}, $writer);
                    $writer->endTag($r_type);
                } elsif ($r_type eq 'phenomLocation') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (keys %{$r->{$r_type}}) {
                        next unless $_ =~ /^i[ns]|from/;
                        $writer->emptyTag($_, 'v', $r->{$r_type}{$_});
                    }
                    _printXMLWeather [$r->{$r_type}{weather}], $writer
                        if exists $r->{$r_type}{weather};
                    if (exists $r->{$r_type}{cloudType}) {
                        $writer->startTag('cloudTypeList');
                        for (@{$r->{$r_type}{cloudType}}) {
                            $writer->emptyTag('cloudType', 'v', $_);
                        }
                        $writer->endTag('cloudTypeList');
                    }
                    $writer->emptyTag('otherPhenom', 'v',
                                      $r->{$r_type}{otherPhenom})
                        if exists $r->{$r_type}{otherPhenom};
                    $writer->emptyTag('cloudCover', 'v',
                                      $r->{$r_type}{cloudCover})
                        if exists $r->{$r_type}{cloudCover};
                    if (exists $r->{$r_type}{lightningType}) {
                        $writer->startTag('lightningTypeList');
                        for (@{$r->{$r_type}{lightningType}}) {
                            $writer->emptyTag('lightningType', 'v', $_);
                        }
                        $writer->endTag('lightningTypeList');
                    }
                    _printXMLLocSpecs $r->{$r_type}, $writer;
                    $writer->endTag($r_type);
                } elsif ($r_type eq 'conditionMountain') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (@{$r->{$r_type}{condMoun}}) {
                        $writer->startTag('condMoun');
                        $writer->emptyTag('condMounType', 'v',
                                          $_->{condMounType});
                        $writer->emptyTag('condMounChange', 'v',
                                          $_->{condMounChange})
                            if exists $_->{condMounChange};
                        _printXMLLocSpecs $_, $writer;
                        $writer->endTag('condMoun');
                    }
                    $writer->endTag($r_type);
                } elsif ($r_type eq 'conditionValley') {
                    $writer->startTag($r_type, 's', $r->{$r_type}{s});
                    for (@{$r->{$r_type}{condVall}}) {
                        $writer->startTag('condVall');
                        $writer->emptyTag('condVallType', 'v',
                                          $_->{condVallType});
                        $writer->emptyTag('condVallChange', 'v',
                                          $_->{condVallChange})
                            if exists $_->{condVallChange};
                        _printXMLLocSpecs $_, $writer;
                        $writer->endTag('condVall');
                    }
                    $writer->endTag($r_type);
                } elsif ($r_type ne 'freetext') {
                    $writer->emptyTag(_mkXMLTag($r, $r_type));
                }
            }
            $writer->endTag('remark');
        }
        $writer->endTag('remarkList');
    }

    _printXMLTag($metar, $writer, 'amendment') if exists $metar->{amendment};

    $writer->endTag($is_taf ? 'taf' : 'metar');
}

sub xml_start {
    my $opts = shift;
    my ($output, $writer);

    $output = new IO::File(">$opts->{x}");
    $writer = new XML::Writer(OUTPUT=>$output, DATA_MODE=>1, DATA_INDENT=>1);
    $writer->xmlDecl("ISO-8859-1");
    if (exists $opts->{X}) {
        $writer->doctype('data', undef, 'metaf.dtd') if defined $opts->{D};
        $writer->pi('xml-stylesheet',
                    'href="' . $opts->{S} . '" type="text/xsl"')
            if defined $opts->{S};
        $writer->startTag('data');
        if (exists $opts->{O}) {
            my %options;
            @{$options{options}}{'type', 'lang', 'format', 'source', 'msg'}
                                                     = split / /, $opts->{O}, 5;
            _printXMLTag(\%options, $writer, 'options');
        }
    }
    $writer->startTag('metafList');
    return $writer;
}

sub xml_end {
    my ($opts, $writer, $versions) = @_;

    $writer->endTag('metafList');
    if (exists $opts->{X}) {
        $writer->startTag('versionList');
        for (@$versions) {
            $writer->dataElement('version', $_);
        }
        $writer->dataElement('version', 'XML: $Id: XML.pm,v 1.11 2006/08/18 09:50:07 metaf2xml Exp $');
        $writer->endTag('versionList');
        $writer->endTag('data');
    }
    $writer->end();
}

1;
__END__
