########################################################################
# metaf2xml/parser.pm -- parse a METAR or TAF message
#
# copyright (c) metaf2xml 2006
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

package metaf2xml::parser;

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;
use vars qw(@ISA @EXPORT @EXPORT_OK);

=head1 NAME

metaf2xml::parser - parse a METAR or TAF message

=head1 SYNOPSIS

  use metaf2xml::parser;

  my %metar = parseReport($msg, $is_taf, $is_speci);

=head1 ARGUMENTS

=over

=item B<$msg>

string that contains the message

=item B<$is_taf>

set if the message is a TAF message

=item B<$is_speci>

set if the message is a SPECI message

=back

=head1 RETURN VALUE

The return value is a hash with all the parsed components of the message. It
may also have values for the following keys:

=over

=item C<ERROR>

If a message could not be parsed C<ERROR> is a hash with the the keys I<descr>
(for where the error occured) and I<pos> - the original message with the
position of the error marked as C<< <@> >>.

=item C<WARNING>

The hash value is a string describing problems which where encountered during
parsing but which didn't prevent complete parsing.

=back

=head1 DESCRIPTION

This module analyzes the string I<$msg> (per default as a METAR message) and
returns a hash with all its components.

=cut

########################################################################
# we need module metaf2xml::basic
# (path may be changed by install!)
########################################################################
use lib '/opt/metaf2xml/lib'; # METAF2XML_LIB
use metaf2xml::basic;

=pod

The module provides 2 functions:

=cut

########################################################################
# export the functions provided by this module
########################################################################
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(
    version_parser
    parseReport
);

########################################################################
# define lots of regular expressions
########################################################################
my $re_day = '(?:[0-2][0-9]|3[01])';
my $re_hour = '(?:[01][0-9]|2[0-3])';
my $re_min = '[0-5][0-9]';
my $re_bbb = '(?:(?:RR|CC|AA)[A-Z]|P[A-Z]{2})';
my $re_rwy = '(?:0[1-9]|[12][0-9]|3[0-6])';
my $re_rwy_des = "$re_rwy(?:[LCR])?";
my $re_rwy_des2 = '(?:[05][1-9]|[1267][0-9]|[38][0-6])'; # add 50 for rrR
my $re_wind_speed = 'P?[1-9]?[0-9]{2}';
my $re_wind_speed_unit = '(?:KT|MPS|KMH)';
my $re_wind_dir = '(?:[0-2][0-9]|3[0-6])';
my $re_wind = "(VRB|///|${re_wind_dir}0)(//|$re_wind_speed)(?:G($re_wind_speed))?($re_wind_speed_unit)";
my $re_vis_m = '[0-9]{4}';
my $re_vis_km = '[1-9][0-9]?KM';
# EXTENSION: allow P6000 (UERP ULOO UMKK UMOO URSS UTAK)
my $re_vis_m_km = "(?:P?$re_vis_m(?:NDV)?|$re_vis_km)";
my $re_vis_m_km_remark = "(?:$re_vis_m|[1-9][0-9]{2,3}M|$re_vis_km)";
my $re_vis_frac16_sm = '[135]/16';
my $re_vis_frac8_sm =  '[1357]/8';
my $re_vis_frac4_sm =  '[13]/4';
my $re_vis_frac2_sm =  '1/2';
my $re_vis_frac_sm0 =
    "(?:$re_vis_frac16_sm|$re_vis_frac8_sm|$re_vis_frac4_sm|$re_vis_frac2_sm)";
my $re_vis_frac_sm1 =
    "1(?: (?:$re_vis_frac8_sm|$re_vis_frac4_sm|$re_vis_frac2_sm))?";
# EXTENSION: allow /8 for 2 miles
my $re_vis_frac_sm2 =
    "2(?: (?:$re_vis_frac8_sm|$re_vis_frac4_sm|$re_vis_frac2_sm))?";
# EXTENSION: allow /2 for 3 miles
my $re_vis_frac_sm3 = "3(?: (?:$re_vis_frac2_sm))?";
my $re_vis_whole_sm = "(?:[04-9]|1[0-5]|[2-9][05]|[1-9][0-9]0)";
my $re_vis_sm =   "(?:M1/4"
                 . "|$re_vis_whole_sm"
                 . "|$re_vis_frac_sm0"
                 . "|$re_vis_frac_sm1"
                 . "|$re_vis_frac_sm2"
                 . "|$re_vis_frac_sm3)";
my $re_rwy_vis = '[PM]?[0-9]{4}';
my $re_rwy_wind = "(?:WIND )?RWY($re_rwy_des) ?((?:///|VRB|${re_wind_dir}0)(?://|$re_wind_speed(?:G$re_wind_speed)?)KT)(?: (${re_wind_dir}[0-9])V(${re_wind_dir}[0-9]))?";

my $re_compass_dir = '(?:[NS]?[EW]|[NS])';
my $re_compass_dir16 = "(?:[NE]NE|[ES]SE|[SW]SW|[NW]NW|$re_compass_dir)";

my $re_phen_locdir_comp = "(?:(?:TO )?(?:GRID )?(?:[1-9][0-9]* ?)?$re_compass_dir16(?: QUAD)?)";
my $re_phen_locdir_comp_thru = "(?:OMTNS|(?:(?:OHD|(?:DSNT |VC )?(?:OBSCG MTNS )?$re_phen_locdir_comp)(?:(?: ?- ?| THRU )$re_phen_locdir_comp|OHD)*))";
my $re_phen_locdir_comp_and = "(?:$re_phen_locdir_comp_thru(?:(?:[/ ]| AND )$re_phen_locdir_comp_thru)*)";
my $re_phen_locdirs = "(?: (?:DSNT |VC )?ALQDS| $re_phen_locdir_comp_and)";
my $re_wx_mov_d = "(?: (MOVD?) ($re_compass_dir16|OHD))";
my $re_num_quadrant1 = '(?:1(?:ST)?|2(?:ND)?|3(?:RD)?|4(?:TH)?)';
my $re_num_quadrant = "(?: $re_num_quadrant1(?:-$re_num_quadrant1)* QUAD)";

my $re_weather_desc = '(?:MI|BC|PR|DR|BL|SH|TS|FZ)';
my $re_weather_ra = 'RA(?:DZ|SN|SG|PL|IC|GR|GS)*';
my $re_weather_dz = 'DZ(?:RA|SN|SG|PL|IC|GR|GS)*';
my $re_weather_sn = 'SN(?:RA|DZ|SG|PL|IC|GR|GS)*';
my $re_weather_ic = 'IC(?:RA|DZ|SN|SG|PL|GR|GS)*';
my $re_weather_prec = '(?:DZ|RA|SN|SG|PL|IC|GR|GS)+';
# EXTENSION: allow SHRADZ (ENHK)
my $re_weather_ts_sh = '(?:RA|SN|PL|GR|GS|DZ)+|UP';
my $re_weather_sg_pl_gr_gs = '(?:SG|PL|GR|GS)+';
# EXTENSION: allow SN, BC, PR, BR, PL with FZ
my $re_weather_fz = '(?:RA|DZ|FG|SN|BC|PR|BR|PL)+';
# EXTENSION: allow +/- with BLSN, BR, TS, FC
my $re_weather_w_i = "(?:$re_weather_prec|DS|SS"
                      . "|(?:TS|SH)(?:$re_weather_ts_sh)"
                      . "|FZ(?:SH)?(?:$re_weather_fz)"
                      . "|UP|BLSN|BR|FC|TS)";
my $re_weather_wo_i = "(?:$re_weather_ic|FG|BR|SA|DU|HZ|FU|VA|SQ|PO|FC|TS"
                       . "|(?:FZ|MI|BC|PR)*FG"
                       . "|BL(?:$re_weather_sn|SA|DU)"
                       . "|DR(?:$re_weather_sn|SA|DU))";
# EXTENSION: allow VCTS..
my $re_weather_vc = "(?:FG|PO|FC|DS|SS|TS(?:$re_weather_ts_sh)?|SH|BL(?:$re_weather_sn|SA|DU)|VA)";
# annex 3, form:
# FZDZ FZRA DZ   RA   SHRA SN SHSN SG SHGR SHGS BLSN SS DS
# TSRA TSSN TSPL TSGR TSGS FC VA   PL UP
# annex 3, text:
# FZDZ FZRA DZ   RA   SHRA SN SHSN SG SHGR SHGS BLSN SS DS
#                          FC VA   PL    TS
# EXTENSION: allow FZFG
my $re_weather_re = "(?:(?:FZ|SH)?$re_weather_prec"
                     . "|TS(?:$re_weather_ts_sh)"
                     . "|FZFG|BLSN|DS|SS|TS|FC|VA|UP)";
# EXTENSION: allow [+-]VC...
my $re_weather = "(?:[+-]?$re_weather_w_i|$re_weather_wo_i|[+-]?VC$re_weather_vc)";

my $re_cloud_cov = '(?:FEW|SCT|BKN|OVC)';
my $re_cloud_base = '(?:///|[0-9]{3})';
my $re_cloud_type = '(?:AC|ACC|ACSL|AS|CB|CBMAM|CC|CCSL|CI|CS|CU|CUFRA'
                     . '|NS|SAC|SC|SCSL|ST|STFRA|TCU|///)';

my $re_colour1 = 'BLU\+?|WHT|GRN|YLO[12]?|AMB|RED';
my $re_colour = "(?:BLACK|(?:BLACK)?(?:$re_colour1)(?: ?(?:$re_colour1|FCST CANCEL))?)";

my $re_be_prec_be   = "(?:[BE]$re_hour?$re_min)";
my $re_be_prec = "(?:(?:FZ|SH)?(?:TS|RA|SN|UP|PL|DZ)+$re_be_prec_be+)";

my $re_rsc_cond = '(?:DRY|WET|SANDED)';
my $re_rsc_deposit = '(?:LSR|SLR|PSR|IR)+(?://|[0-9]{2})?';
my $re_rsc = "(?:$re_rsc_deposit(?:P(?:[ /]$re_rsc_cond)?)?|(?:(?:P[ /])?$re_rsc_cond))";

my $re_opac_cloud =
                  '(?:(?:BL)?SN|FG|AS|AC|CC|CF|CI|CS|CU|TCU|NS|ST|SF|SC|ACC|CUFRA|CB)';

my $re_phen_desc =
               '(?:LOW|OCNL|FRQ|(?:VRY )?LGT|INTMT|CONS|PAST HR|ISOL|CONTRAIL)';
my $re_ltg_types = '(?:CA|CC|CG|CW|IC)';

my $re_wind_shear_lvl = "WS([0-9]{3})/(${re_wind_dir}0${re_wind_speed}KT)";

my %sea_cond = (
    0 => 'calm (glassy)',
    1 => 'calm (rippled)',
    2 => 'smooth (wavelets)',
    3 => 'slight',
    4 => 'moderate',
    5 => 'rough',
    6 => 'very rough',
    7 => 'high',
    8 => 'very high',
    9 => 'phenomenal',
   '/' => '(not available)',
);

my %sea_swell = (
    0 => 'none',
    1 => 'Length: short or average, height: low',
    2 => 'smooth (wavelets)',
    3 => 'slight',
    4 => 'moderate',
    5 => 'rough',
    6 => 'very rough',
    7 => 'high',
    8 => 'very high',
    9 => 'phenomenal',
   '/' => '(not available)',
);

$ENV{TZ} = 'GMT0';

########################################################################
# version_parser
########################################################################

=over

=item version_parser()

It returns the CVS (RCS) id of the module file.

=cut

sub version_parser {
    return 'parser: $Id: parser.pm,v 1.19 2006/08/18 09:51:40 metaf2xml Exp $';
}

sub _printCldWOpacity {
    my $clds = shift;
    my ($txt, $hdr);

    $hdr = "Cloud with opacity:\t";
    for ($clds =~ /((?:[A-Z]+[0-9])+?)/g) {
        /(.*)(.)/;
        $txt .= $hdr
               . "$2/8 "
               . (index($1, 'SN') >-1 || index($1, 'FG') > -1 ? printWeather(_parseWeather($1), 1) : $cloud_types{$1});
        $hdr = "\n\t\t\t";
    }
    return $txt;
}

# $cy: [ 'ABCD', 'AB', 'A' ]
sub _isValidCountry {
    my ($re, $cy) = @_;

    my %country = (
        cloudMaxCover =>
            ' LO ',
        NEFO_PLAYA =>
            ' SCSE ',
        rwyWind =>
            ' LG ',
        remarkCloudMaxCov =>
            ' LI OA ',
        remarkColour =>
            ' EG ET FH L OA ',
        remarkCloudType =>
            ' K MM MU NS P RJ RO UT ',
        remarkCIGRAG =>
            ' CY P K ',
        remarkCIGVRBL =>
            ' CY ',
        remarkSLPdPa =>
            ' BGTL CM CW CY CZ ET K LQ MM MU NS NZ P RJ RK RO TI TJ TX ',
        remarkSLPinHg =>
            ' ',
        remarkGRIDWind =>
            ' NZ ',
        remarkRwyWind =>
            ' BG LC LT SP ',
        remarkRsc =>
            ' BG ET K NZ PA ',
        remarkCloudOpacity =>
            ' CQ CW CY CZ ',
        remarkCloudOpacityRev =>
            ' BG KTTS ',
        remarkCloudOpacityLvl =>
            ' RJ YB YM YP ',
        remarkCloudCoverLvl =>
            ' C EK K LI LL P NS TI TJ ',
        remarkCloudTypeLvl =>
            ' U ',
        remarkReWeather =>
            ' CY FQ TR ',
        remarkWR =>
            ' EG ET K LQ P R ',
        remarkSfcVisM =>
            ' KQ ',
        remarkVisSfcTwr =>
            ' K PA ',
        remarkVisVarK =>
            ' K PA ',
        remarkVisVarCY =>
            ' CY ',
        remarkPCPNLastHour =>
            ' CW CY CZ ',
        remarkVisLocM =>
            ' E KQ LQ RJ ',
        remarkVisLwr =>
            ' CY K PA ',
        remarkCigLwr =>
            ' CY K PA ',
        remarkQNH =>
            ' RJ RP ',
        remarkQFF =>
            ' MM ',
        remarkCorrected =>
            ' ET K LQ PA ',
        remarkSnowIncr =>
            ' K PA ',
        remarkHourly =>
            ' BG C ET K LQ MM NS P TI TJ ',
        remarkPeakWind =>
            ' C ET K LQ NZ PA PH TI TJ ',
        remarkHourlyTemp =>
            ' C K P NS MU RJ TI TJ ',
        remarkQFEhPa =>
            ' UTTP UTSS ',
        remarkDA_PA =>
            ' K CY ',
        remarkKN =>
            ' KN ',
        remarkKNIP =>
            ' KNIP ',
        remarkQFEmmHg =>
            ' U ZM ',
        remarkPhenCloud =>
            ' K PA PH ',
        remarkCloudCoverVar =>
            ' K MH PA PH PP ',
        remarkC_K_P_NS_TI_TJ_ET =>
            ' C K P NS TI TJ ET ',
        remarkSNWCVR =>
            ' C ',
        remarkObsTaken =>
            ' C ',
        remarkBalloon =>
            ' CY CZ ',
        remarkRVR =>
            ' CY ',
        remarkTAFNxtFcstBy =>
            ' CY CZ ',
        remarkTAFNxtFcstAt =>
            ' CY ',
        remarkTAFFcstAutoObs =>
            ' CY CZ ',
        remarkTAFFcstAutometar =>
            ' EF ',
    );
    return    index($country{$re}, ' ' . $cy->[0] . ' ') > -1
           || index($country{$re}, ' ' . $cy->[1] . ' ') > -1
           || index($country{$re}, ' ' . $cy->[2] . ' ') > -1;
}

sub _parseSM {
    my $vis = shift;

    return $vis if $vis =~ /^$re_vis_whole_sm$/o;
    return $vis if $vis =~ '[.]';

    if ($vis =~ /^$re_vis_frac_sm0$/o) {
        my ($f1, $f2) = $vis =~ '([0-9]*)\/([0-9]*)';
        return $f1 / $f2;
    }

    my ($whole, $f1, $f2) = $vis =~ '(.)(?: ([0-9]*)\/([0-9]*))?';
    $whole += $f1 / $f2 if $f2;
    return $whole;
}

sub _parseWeather {
    my ($weather, $is_recent) = @_;
    my ($w, $int, $vc, $desc, $phen, $intensity, $delim, $weather_str);

    $weather_str = ($is_recent ? 'RE' : '') . $weather;
    return { notAvailable => 1, s => $weather_str } if $weather eq '//';
    return { NSW => 1, s => $weather_str } if $weather eq 'NSW';

    return { tornado => 1, s => $weather_str } if $weather eq '+FC';

    ($int, $vc, $desc, $phen) =
                      $weather =~ m@([+-])?(VC)?($re_weather_desc+)?([A-Z/]+)@o;
    $w->{s}          = $weather_str;
    $w->{intensity}  = ($int eq '-' ? 'LIGHT' : 'HEAVY') if defined $int;
    $w->{inVicinity} = 1 if defined $vc;
    @{$w->{descriptor}} = $desc =~ /(..)+?/g if defined $desc;
    @{$w->{phenomenon}} = $phen =~ /(..)+?/g if defined $phen;
    return $w;
}

sub _parseRwyState {
    my ($metar, $msg) = @_;
    my $r;

    if ($$msg =~ m@^((SNOCLO)|(88|99|$re_rwy_des2)(?:(CLRD)|([0-9/])([0-9/])([0-9]{2}|//))([0-9]{2}|//)) (.*)@o)
    {
        $$msg = $9;
        $r->{s} = $1;
        if (defined $2) {
            $r->{SNOCLO} = 1;
        } else {
            if ($3 == 88) {
                $r->{rwyDesigAll} = 1;
            } elsif ($3 == 99) {
                $r->{rwyDesigRep} = 1;
            } elsif ($3 > 50) {
                $r->{rwyDesig} = sprintf '%02dR', $3 - 50;
            } else {
                $r->{rwyDesig} = $3;
            }
            if (defined $4) {
                $r->{cleared} = 1;
            } else {
                if ($5 eq '/') {
                    $r->{depositType}{notAvailable} = 1;
                } else {
                    $r->{depositType}{depositTypeVal} = $5;
                }
                if ($6 eq '/') {
                    $r->{depositExtent}{notAvailable} = 1;
                } elsif (   $6 eq '0' || $6 eq '1' || $6 eq '2'
                         || $6 eq '5' || $6 eq '9')
                {
                    $r->{depositExtent}{depositExtentVal} = $6;
                } else {
                    $r->{depositExtent}{invalidFormat} = $6;
                }
                if ($7 eq '//') {
                    $r->{depositDepth}{notAvailable} = 1;
                } elsif ($7 != 91) {
                    $r->{depositDepth}{depositDepthVal} = $7 + 0;
                } else {
                    $r->{depositDepth}{invalidFormat} = $7;
                }
            }
            if ($8 eq '//') {
                $r->{friction}{notAvailable} = 1;
            } elsif ($8 >= 1 && $8 <= 90) {
                $r->{friction}{coefficient} = $8 + 0;
            } elsif ($8 == 99) {
                $r->{friction}{unreliable} = 1;
            } elsif ($8 >= 91 && $8 <= 95) {
                $r->{friction}{brakingAction} = $8;
            } else {
                $r->{friction}{invalidFormat} = $8;
            }
        }
        push @{$metar->{rwyState}}, $r;
        return 1;
    }
    return 0;
}

sub _parseWind {
    my $wind = shift;
    my ($w, $dir, $speed, $gustSpeed, $unitSpeed);

    return { notAvailable => 1 } if $wind eq '/////';
    ($dir, $speed, $gustSpeed, $unitSpeed) = $wind =~ m@^$re_wind$@o;
    if ($dir eq '///' && $speed eq '//') {
        $w->{notAvailable} = 1;
    } else {
        if ($dir eq '///') {
            $w->{dirNotAvailable} = 1;
        } elsif ($dir eq 'VRB') {
            $w->{dirVariable} = 1;
        } else {
            $w->{dir} = $dir + 0;
        }
        if ($speed eq '//') {
            $w->{speedNotAvailable} = 1;
        } else {
            $w->{speedGreater} = 1 if $speed =~ s/^P//;
            $w->{speed} = $speed + 0;
            $w->{unitSpeed} = $unitSpeed;
        }
        if (defined $gustSpeed) {
            $w->{gustSpeedGreater} = 1 if $gustSpeed =~ s/^P//;
            $w->{gustSpeed} = $gustSpeed + 0;
        }
    }
    return $w;
}

sub _parseCloud {
    my $cloud = shift;
    my $c;

    $c->{s} = $cloud;
    if ($cloud eq '//////') {
        $c->{notAvailable} = 1;
    } elsif ($cloud =~ m@^$re_cloud_cov$re_cloud_base$re_cloud_type?@o){
        $cloud =~ '(...)(?:(...)(.+)?)?';
        $c->{cloudCover} = $1;
        if (defined $2) {
            if ($2 eq '///') {
                $c->{cloudBaseNotAvailable} = 1;
            } else {
                $c->{cloudBase} = $2 + 0;
            }
        }
        if (defined $3) {
            if ($3 eq '///') {
                $c->{cloudTypeNotAvailable} = 1;
            } else {
                $c->{cloudType} = $3;
            }
        }
    } else {
        $cloud =~ '(...)(.+)?';
        $c->{cloudCover} = $1;
        $c->{cloudType}  = $2 if defined $2;
    }
    return $c;
}

sub _parseColourCode {
    my $colour = shift;
    my $c;

    $colour =~ "^(BLACK)?($re_colour1)? ?(.+)?";
    $c->{s} = $colour;
    $c->{BLACK} = 1 if defined $1;
    $c->{currentColour} =
            $2 eq 'BLU+' ? 'BLUplus' : ($2 eq 'FCST CANCEL' ? 'FCSTCANCEL' : $2)
        if defined $2;
    $c->{predictedColour} =
            $3 eq 'BLU+' ? 'BLUplus' : ($3 eq 'FCST CANCEL' ? 'FCSTCANCEL' : $3)
        if defined $3;
    return $c;
}

sub _parseLocations {
    my $loc_str = shift;

    if ($loc_str =~ m@(?:(DSNT|VC) )?(ALQDS)@) {
        my $l;

        $l->{isDistant}    = 1 if defined $1 and $1 eq 'DSNT';
        $l->{inVicinity}   = 1 if defined $1 and $1 eq 'VC';
        $l->{locationSpec} = $2;
        return [ [ $l ] ];
    }

    my $loc_and;
    for ($loc_str =~ m@(?:(?:[/ ]| AND )?($re_phen_locdir_comp_thru)+)@og) {
        my $loc_thru;
        for ($_ =~ m@(?: ?[/-] ?| THRU )?(OHD|OMTNS|(?:DSNT |VC )?(?:OBSCG MTNS )?$re_phen_locdir_comp)@og)
        {
            my $l;

            if ($_ =~ 'OHD|OMTNS') {
                $l->{locationSpec} = $_;
            } else {
                $_ =~ m@(?:(DSNT|VC) )?(OBSCG MTNS )?(?:TO )?(?:(GRID )?(?:([1-9][0-9]*) ?)?($re_compass_dir16)(?: (QUAD))?)@o;
                $l->{isDistant}  = 1 if defined $1 and $1 eq 'DSNT';
                $l->{inVicinity} = 1 if defined $1 and $1 eq 'VC';
                $l->{obscgMtns}  = 1 if defined $2;
                $l->{isGrid}     = 1 if defined $3;
                if (defined $4) {
                    $l->{distance}   = $4;
                    $l->{unitLength} = 'SM';
                }
                $l->{compassDir} = $5;
                $l->{isQuadrant} = 1 if defined $6;
            }
            push @{$loc_thru}, $l;
        }
        push @{$loc_and}, $loc_thru;
    }
    return $loc_and;
}

sub _parseNotRec {
    my $s = shift;
    my ($delim, $delim2, $txt, $re_not_rec);

    $re_not_rec = "(?:(?:FIRST|LAST|BA (?:GOOD|POOR)|THN SPTS IOVC|(?:WND(?: DATA)?|CIG|SLP|ALSTG(?:/SLP)?) ESTMD|ESTMD SLP|FUOCTY|ALTM MISG|ALL WNDS GRID|CVCTV CLDS? EMBD|CONTRAILS?|SKY OBSCURED|FOGGY|BBLO) )";

    return "NOT RECOGNIZED:\t\t$s"
        unless ($s . ' ') =~ m@^$re_not_rec+$@o;

    $delim2 = '';
    for (($s . ' ') =~ m@$re_not_rec@og) {
        $txt .= $delim2;
        $delim2 = "\n";
        $delim = '';
        for (m@(\S+)@g) {
            if ($_ eq 'ALSTG/SLP') {
                $txt .= $delim . $keywords{ALSTG} . '/' . $keywords{SLP};
            } elsif ($_ =~ m@^(?:$re_weather)$@o) {
                $txt .= $delim . printWeather(_parseWeather $_);
            } elsif ($_ =~ m@^$re_cloud_type$@o) {
                $txt .= $delim . $cloud_types{$_};
            } elsif (exists $keywords{$_}) {
                $txt .= $delim . $keywords{$_};
            } else {
                $txt .= $delim . $_;
            }
            $txt .= ' observation' if $_ =~ 'LAST|FIRST';
            $delim = ' ';
        }
    }
    return "$txt";
}

sub _determineCeiling {
    my $cloud = shift;
    my ($ceil, $idx, $ii);
    $ceil = 1000;
    $idx = -1;
    $ii = -1;
    for (@$cloud) {
        $ii++;
        if (   exists $_->{cloudBase}
            && $_->{cloudBase} < $ceil
            && $_->{cloudBase} < 200 # TODO: actually MSL!
            && ($_->{cloudCover} eq 'BKN' || $_->{cloudCover} eq 'OVC'))
        {
            $ceil = $_->{cloudBase};
            $idx = $ii;
        }
    }
    $cloud->[$idx]{isCeiling} = 1 if $idx > -1;
}

sub _parseQuadrants {
    my ($q, $isDistant) = @_;
    my $l;

    $l->{isDistant} = 1 if $isDistant;
    @{$l->{quadrant}} = $q =~ /-?([1-4])[A-Z]*/g;

    return [[$l]];
}

sub _parseUSTemp {
    my $temp = shift;

    $temp =~ /(.)(.*)/;
    return sprintf "%.1f", ($1 eq '1' ? -1 : 1) * $2 / 10;
}

sub _parsePhenomDescr {
    my ($r, $phen_descr) = @_;

    $r->{{
          FRQ       => 'isFrequent',
          OCNL      => 'isOccasional',
          INTMT     => 'isIntermittent',
          CONS      => 'isContinuous',
          LGT       => 'isLight',
         'VRY LGT'  => 'isVeryLight',
          LOW       => 'isLow',
          ISOL      => 'isIsolated',
          CONTRAIL  => 'fromContrails',
         'PAST HR'  => 'inPastHour',
         }->{$phen_descr} } = 1;
}

# "supplementary" section of TAFs
sub _parseTAFsuppl {
    my ($metar, $msg, $base_metar) = @_;
    my $r;

    if ($$msg =~ m@^(([56])([0-9])([0-9]{3})([0-9])) (.*)@o) {
        my ($ti, $type);

        $$msg = $6;
        $type = $2 eq '5' ? 'turbulence' : 'icing';
        $r->{$type} = {
            s               => $1,
            $type . 'Descr' => $3,
            layerBase       => $4 + 0,
        };
        $ti = $5;

        # if item repeated: level is layer top
        if ($$msg =~ m@^($2$3([0-9]{3})$5) (.*)@) {
            $$msg = $3;
            $r->{$type}{s} .= ' ' . $1;
            $r->{$type}{layerTop} = $2 + 0;
        } else {
            $r->{$type}{layerThickness} = $ti;
        }
        push @{$metar->{TAFsuppl}}, $r;
        return 1;
    }

    if ($$msg =~ m@^($re_wind_shear_lvl) (.*)@o) {
        $$msg = $4;
        $r->{windShearLvl} = {
            s     => $1,
            level => $2 + 0,
            wind  => _parseWind $3,
        };
        push @{$metar->{TAFsuppl}}, $r;
        return 1;
    }

    if ($$msg =~ m@^(T(M)?([0-9]{2})/($re_hour)Z) (.*)@o) {
        $$msg = $5;
        $r->{s} = $1;
        $r->{temp} = $3 + 0;
        $r->{temp} *= -1 if defined $2;
        $r->{tempUnit} = 'C';
        $r->{hour} = $4;
        push @{$base_metar->{TAFsuppl}}, { tempAt => $r };
        return 1;
    }

    return 0;
}


########################################################################
# parseReport
########################################################################

=item parseReport()

This is the main function of the module. It parses the message string in
the following steps:

=cut

sub parseReport {
    my ($msg, $is_taf, $is_speci) = @_;

    my (%metar, $msg_hdr);
    my ($day, $hour, $minute);
    my ($curr_day, $curr_hour, $curr_minute);
    my ($ii, $td, @cy, $delim, $tt);

=pod

Leading and trailing spaces are removed, multiple spaces are replaced by a
single one.

=cut

    $msg =~ s/ +/ /g;
    $msg =~ s/ $//;
    $msg =~ s/^ //;

    $metar{msg} = $msg;

    $msg_hdr = '';
    if ($msg =~ s/^(METAR )//) {
        $is_taf = 0;
        $is_speci = 0;
        $msg_hdr = $1;
    } elsif ($msg =~ s/^(SPECI )//) {
        $is_taf = 0;
        $is_speci = 1;
        $msg_hdr = $1;
    } elsif ($msg =~ s/^(TAF )//) {
        $is_taf = 1;
        $msg_hdr = $1;
    }

    $metar{SPECI} = 1 if $is_speci;

    $metar{WARNING} = '';

    $msg .= ' '; # this makes parsing much easier

=pod

The message is checked for typical errors and corrected. Errors can be:

=over 2

=item

invalid characters (everything except capital letters, digits, dot, slash,
dollar sign, space, plus sign, minus sign)

=item

QNH with spaces

=item

temperature or dew point with spaces

=item

misspelled key words, or with missing or additional spaces

=item

missing key words

=item

wrong order of keywords

=item

removal of slashes before and after some components

=back

If the message is changed there will be a C<WARNING>.

=cut

    # EXTENSION: preprocessing
    # invalid characters
    $msg =~ s/[^A-Z0-9.\/\$ +-]/?/gos;

    # QNH with spaces
    $msg =~ s/(?<= A) (?=[23][0-9]{3} )//o;
    $msg =~ s/(?<= Q) (?=[01][0-9]{3} )//o;
    $msg =~ s/(?<= (?:A[23]|Q[01])) (?=[0-9]{3} )//o;
    $msg =~ s/(?<= (?:A[23]|Q[01])[0-9]) (?=[0-9]{2} )//o;
    $msg =~ s/(?<= (?:A[23]|Q[01])[0-9]{2}) (?=[0-9] )//o;

    # temperature or dew point with spaces
    $msg =~ s@( M?[0-9]) (?=[0-9]/M?[0-9]{2} )@$1@o;
    $msg =~ s@( M?[0-9]{2}/M?) (?=[0-9]{2} )@$1@o;
    $msg =~ s@( M?[0-9]{2}/M?[0-9]) (?=[0-9] )@$1@o;

    # misspelled key words, or with missing or additional spaces
    $msg =~ s/ (?:CAMOK|CC?VOK) / CAVOK /o;
    $msg =~ s/ (?:NO(?: S)?IG|NOSI(?: G)?|N[L ]OSIG|NSG|(?:MO|N0)SIG|NOS I ?G) / NOSIG /o;
    $msg =~ s/(?<! )(?=(?:TEMPO|BECMG) )/ /o; # BLU+BLU+(TEMPO|BECMG)
    $msg =~ s/ (?:R MK|RMKS|RRMK) / RMK /o;
    $msg =~ s@(?<= RMK)(?=[^ ])@ @o;
    $msg =~ s@ 0VC(?=$re_cloud_base$re_cloud_type?|$re_cloud_type )@ OVC@og;
    $msg =~ s@ BNK(?=$re_cloud_base$re_cloud_type?|$re_cloud_type )@ BKN@og;
    $msg =~ s@( $re_cloud_type| FULYR)[/-](?=$re_compass_dir16 )@$1 @og;
    $msg =~ s@(?<= VR) (?=B$re_wind_speed$re_wind_speed_unit )@@og;
    $msg =~ s@(?<=[0-9]KT)S(?= )@@og;
    $msg =~ s@(?<= R) (?=WY)@@og;
    $msg =~ s@(?<= RW) (?=Y)@@og;
    $msg =~ s@(?<= RWY$re_rwy) (?=[LCR] )@@og;
    $msg =~ s@ RNW @ RWY @og;
    $msg =~ s@/(?=AURBO )@ @o;
    $msg =~ s@(?<= A)0(?=[12] )@O@o;
    $msg =~ s@(?<= ALQ)(?=S )@D@o;
    $msg =~ s@(?<= O)V(?=HD )@@o;
    $msg =~ s@(?<= O)VH(?= )@HD@o;
    $msg =~ s@(?<= EST)(?= )@MD@o;
    $msg =~ s@(?<= EST)(?=D )@M@o;
    $msg =~ s@(?<= ES)MT(?=D )@TM@o;
    $msg =~ s@(?<= OCNL)Y(?= )@@o;
    $msg =~ s@(?<= PK)(?=WND )@ @o;
    $msg =~ s@( $re_phen_desc L)GT(?=$re_ltg_types+)@$1TG@o;
    $msg =~ s@( $re_phen_desc LTG) (?=$re_ltg_types+)@$1@o;
    $msg =~ s@ LIGHTNING @ LTG @og;
    $msg =~ s@(?<= I)N(?=OVC )@@o;
    $msg =~ s@(?<= TH)I(?=N )@@o;
    $msg =~ s@(?<= SP)O(?=TS )@@o;
    $msg =~ s@(?<= W) (?=IND )@@o;
    $msg =~ s@(?<= WIND RWY) (?=$re_rwy_des )@@o;
    $msg =~ s@(?<= (?:RWY|THR))($re_rwy_des ${re_wind_dir}0$re_wind_speed) (?=$re_wind_speed_unit )@$1@og;
    $msg =~ s@(?<= S)OUTH(?= )@@o;
    $msg =~ s@(?<= B)A(?=NK )@@o;
    $msg =~ s@(?<= F)(?= BNK )@G@o;
    $msg =~ s@( [+-]?)(RA|SN)SH @$1SH$2 @go;
    $msg =~ s@(?<= ISOL)D(?= )@@o;
    $msg =~ s@(?<= H)A(?=ZY )@@o;
    $msg =~ s@(?<= DS)(?=T )@N@o;
    $msg =~ s@(?<= D)IS(?=T )@SN@o;
    $msg =~ s@ ESTIMATED PASS OPEN @ EPO @o;
    $msg =~ s@ PASS ESTD CLSD @ EPC @o;
    $msg =~ s@ OVR MT @ OMTNS @o;
    $msg =~ s@(?<= PL)(?=ME )@U@o;
    $msg =~ s@(?<= BL)K @ACK@o;
    $msg =~ s@(?<= CON)TU(?=S )@@o;
    $msg =~ s@(?<= INVIS)T(?= )@@o;
    $msg =~ s@(?<= FROIN)/ ?@ @o;
    $msg =~ s@(?<= AS)(?=CTD )@O@o;
    $msg =~ s@(?<=^[A-Z][A-Z0-9]{3} $re_day$re_hour$re_min)(?= )@Z@o;
    $msg =~ s@(?<=^[A-Z][A-Z0-9]{3} $re_day$re_hour$re_min)KT(?= )@Z@o;
    $msg =~ s@(?<=^[A-Z][A-Z0-9]{3} $re_day$re_hour${re_min}Z )(${re_wind_dir}0$re_wind_speed)TK(?= )@$1KT@o;
    $msg =~ s@(?<=^[A-Z][A-Z0-9]{3} $re_day$re_hour${re_min}Z )(${re_wind_dir}0)/(?=${re_wind_speed}$re_wind_speed_unit )@$1@o;
    $msg =~ s@(?<=M)/(?=S )@P@o;
    $msg =~ s@(?<=[+]) (?=$re_weather_w_i )@@o;
    $msg =~ s@(?<=$re_cloud_base )(M?[0-9]{2}/)/(?=M?[0-9]{2} )@$1@o;
    $msg =~ s@(?<= BECMG| TEMPO)(?=$re_hour(?:$re_hour|24) )@ @go;

    # missing key words
    $msg =~ s@ (?:TEMPO|INTER) ($re_hour$re_min)/($re_hour$re_min)(?= )@ TEMPO FM$1 TL$2@og;
    $msg =~ s@(?<! (?:BECMG|TEMPO) )(?=FM$re_hour$re_min )@BECMG @og
        unless $is_taf;

    # wrong order of keywords
    $msg =~ s@( DSNT)( $re_weather| $re_cloud_type| LTG)(?=$re_phen_locdirs )@$2$1@og;
    $msg =~ s@( $re_weather| $re_cloud_type| LTG)($re_phen_locdirs|$re_num_quadrant) (DSNT) @$1 $3$2 @og;

    # KNPA: PA, DA
    $msg =~ s@ /(PA ?[+-]?[0-9]+ DA ?[+-]?[0-9]+)/ @ $1 @;

    $metar{msgCorr} = $msg;
    if ($metar{msg} . ' ' ne $msg_hdr . $msg) {
        # we appended a blank, don't show it in warning
        my $msg_mod;
        $msg_mod = "$msg_hdr$msg";
        $msg_mod =~ s/ $//;
        $metar{WARNING} .= "WARNING: message modified to: $msg_mod\n"
    }

=pod

Then the components of the message are checked starting with the observation
station and the observation time (or issue time for TAF).

=cut

    if ($msg !~ /^([A-Z][A-Z0-9]{3}) (.*)/o) {
        $metar{ERROR}{descr} = 'obsStation';
        $metar{ERROR}{pos}   = substr($metar{msgCorr}, 0,
                                      (length $metar{msgCorr}) - length $msg)
                               . '@> ' . $msg;
        return %metar;
    }
    $metar{obsStation}{id} = $1;
    $metar{obsStation}{s} = $1;
    $msg = $2;

    @cy = $metar{obsStation}{id} =~ /(((.).)..)/;

    $tt = $is_taf ? 'issueTime' : 'obsTime';
    if ($msg =~ /^($re_day)($re_hour)($re_min)Z (.*)/o) {
        $msg = $4;
        $metar{$tt}{s} = "$1$2$3Z";
        @{$metar{$tt}}{'day', 'hour', 'minute'} = ($1, $2, $3);
    } elsif ($msg =~ /^($re_day$re_hour${re_min}KT) (.*)/o) {
        $msg = $2;
        $metar{$tt}{s} = $1;
        $metar{$tt}{invalidFormat} = $1;
    } elsif ($msg =~ /^(NIL) $/o) {
        $metar{reportModifier}{modifierType} = $metar{reportModifier}{s} = $1;
        return %metar;
    } else {
        $metar{ERROR}{descr} = 'obsTime';
        $metar{ERROR}{pos}   = substr($metar{msgCorr}, 0,
                                      (length $metar{msgCorr}) - length $msg)
                               . '@> ' . $msg;
        return %metar;
    }

=pod

For METAR messages the next component may optionally be the report modifier to
indicate a missing message (C<NIL>), a message created by an automated station
(C<AUTO>), or an corrected (C<COR>) or retarded (C<RTD>) message. For Canada,
the component BBB may appear to indicate that the report has been retarded
(C<RR?>), corrected (C<CC?>), amended (C<AA?>), or segmented (C<P??>).

=cut

    # EXTENSION: BBB (for Canada)
    if (!$is_taf && $msg =~ /^(NIL|AUTO|COR|RTD|$re_bbb) (.*)/o) {
        $msg = $2;
        $metar{reportModifier}{s} = $1;
        if ($metar{reportModifier}{s} =~ 'NIL|AUTO|COR|RTD') {
            $metar{reportModifier}{modifierType} = $metar{reportModifier}{s};
        } else {
            $metar{reportModifier}{s} =~ '(.)(.)(.)';
            if ($1 eq 'P') {
                $metar{reportModifier}{modifierType} = $1;
                $metar{reportModifier}{segment} = "$2$3";
                $metar{reportModifier}{isLastSegment} = 1 if $2 eq 'Z';
            } else {
                $metar{reportModifier}{modifierType} = "$1$2";
                if ($3 eq 'Z') {
                    $metar{reportModifier}{over24hLate}  = 1;
                } elsif ($3 eq 'Y') {
                    $metar{reportModifier}{sequenceLost} = 1;
                } else {
                    $metar{reportModifier}{bulletinSeq} = $3;
                }
            }
        }
        return %metar if $metar{reportModifier}{modifierType} eq 'NIL';
    }

=pod

For TAF messages there may follow the modifier C<NIL> (which terminates the
message), C<COR> (corrected), or C<AMD> (amended). Then follows the forecast
period and optionally C<CNL> to indicate a cancelled forecast.

=cut

    if ($is_taf) {
        if ($msg =~ /^(NIL|COR|AMD) (.*)/o) {
            $msg = $2;
            $metar{reportModifier}{s} = $1;
            $metar{reportModifier}{modifierType} = $1;
            return %metar if $1 eq 'NIL';
        }

        if ($msg =~ m/^($re_day)($re_hour)($re_hour|24) (.*)/o) {
            $msg = $4;
            $metar{fcstPeriod}{s} = "$1$2$3";
            @{$metar{fcstPeriod}}{'day', 'hourFrom', 'hourTill'} = ($1, $2, $3);
        } else {
            $metar{ERROR}{descr} = 'fcstPeriod';
            $metar{ERROR}{pos}   = substr($metar{msgCorr}, 0,
                                          (length $metar{msgCorr}) - length $msg)
                                   . '@> ' . $msg;
            return %metar;
        }

        if ($msg =~ m/^(CNL) (.*)/o) {
            $msg = $2;
            $metar{fcstCancelled}{s} = "$1";
            return %metar;
        }
    }

=pod

The first weather group is the surface wind. It may include a gust speed; the
wind speed and/or the wind direction may also be missing. There may be a second
group to report a varying wind direction.

=cut

    # EXTENSION: allow ///// (wind not available)
    # EXTENSION: allow /// for wind direction (not available)
    # EXTENSION: allow // for wind speed (not available)
    if ($msg =~ m@^(/////|(?:$re_wind)) (.*)@o)
    {
        $msg = $6;
        $metar{sfcWind}{s} = $1;
        $metar{sfcWind}{wind} = _parseWind $1;
    # EXTENSION: allow invalid formats
    } elsif ($msg =~ m@^(${re_wind_dir}0$re_wind_speed|[0-9]{6}$re_wind_speed_unit) (.*)@o) {
        $metar{sfcWind}{wind}{invalidFormat} = $1;
        $metar{sfcWind}{s} = $1;
        $msg = $2;
    # EXTENSION: allow missing wind
    } else {
        $metar{WARNING} .= "WARNING: No wind\n";
    }

    # EXTENSION:
    # Annex 3 Appendix 3 4.1.4.2.b.1 requires wind speed >=3 kt and
    #   variation 60-180 for this group: not checked
    if ($msg =~ /^(${re_wind_dir}[0-9])V(${re_wind_dir}[0-9]) (.*)/o) {
        if (exists $metar{sfcWind}) {
            $metar{sfcWind}{s} .= " ";
        } else {
            # TODO?: actually not "not available" but missing
            $metar{sfcWind}{wind}{dirNotAvailable} = 1;
            $metar{sfcWind}{wind}{speedNotAvailable} = 1;
            $metar{sfcWind}{s} = '';
        }
        $metar{sfcWind}{wind}{windVarLeft} = $1 + 0;
        $metar{sfcWind}{wind}{windVarRight} = $2 + 0;
        $metar{sfcWind}{s} .= "$1V$2";
        $msg = $3;
    }

=pod

If the weather conditions allow (visbility more than 10 km, no relevant weather,
no clouds below 5000 ft) this may be indicated by the keyword C<CAVOK>. The next
component after that should be the temperature (see below).

=cut

    # Annex 3 Appendix 3 2.2:
    #   vis >10km, no relevant weather, no cloud below 5000 ft
    if ($msg =~ '^CAVOK (.*)') {
        $metar{CAVOK} = 1; # cancels visibility, weather and cloud
        $msg = $1;
    }

    if (!$metar{CAVOK}) {

=pod

Otherwise follow the prevailing visibilty. It can have a compass direction
attached or C<NDV> if no directional variations can be given. There may be an
additional group for the minimum visibilty.

=cut

        # EXTENSION: allow 16 compass directions
        # EXTENSION: allow 'M' after re_vis_m
        if ($msg =~ /^((?:(P)?($re_vis_m)M?(NDV)?|($re_vis_km))($re_compass_dir16)?)(?: ($re_vis_m)($re_compass_dir))? (.*)/o)
        {
            $msg = $9;
            $metar{visPrev}{s} = $1;
            $metar{visPrev}{compassDir} = $6 if defined $6;
            if ($7) {
                $metar{visMin}{s}          = $7 . $8;
                $metar{visMin}{distance}   = $7 + 0;
                $metar{visMin}{unitLength} = 'M';
                $metar{visMin}{compassDir} = $8;
            }
            if (defined $3) {
                $metar{visPrev}{distance}       = $3 + 0;
                $metar{visPrev}{unitLength}     = 'M';
                $metar{visPrev}{isEqualGreater} = 1 if defined $2;
                $metar{visPrev}{NDV}            = 1 if defined $4;
            } else {
                $metar{visPrev}{distance}   = $5;
                $metar{visPrev}{distance}   =~ s/KM//;
                $metar{visPrev}{unitLength} = 'KM';
            }
        } elsif ($msg =~ m@^(${re_vis_sm})SM (.*)@o) {
            $msg = $2;
            $metar{visPrev}{s} = "$1SM";
            my ($is_less, $vis) = $1 =~ '(M)?(.*)';
            $metar{visPrev}{distance}   = _parseSM $vis;
            $metar{visPrev}{unitLength} = 'SM';
            $metar{visPrev}{isLess}     = 1 if defined $is_less;
        } elsif ($is_taf && $msg =~ m@^P([1-9][0-9]*)SM (.*)@o) {
            $msg = $2;
            $metar{visPrev}{s} = "P$1SM";
            $metar{visPrev}{distance}       = $1;
            $metar{visPrev}{unitLength}     = 'SM';
            $metar{visPrev}{isEqualGreater} = 1;
        # EXTENSION: allow //// and misformatted entry
        } elsif ($msg =~ m@^//// (.*)@o) {
            $metar{visPrev}{notAvailable} = 1;
            $metar{visPrev}{s} = '////';
            $msg = $1;
        } elsif ($msg =~ m@^([0-9]{3})(?: ($re_vis_m)($re_compass_dir))? (.*)@o)
        {
            $metar{visPrev}{invalidFormat} = $1;
            $metar{visPrev}{s} = $1;
            if ($2) {
                $metar{visMin}{s}          = $2 . $3;
                $metar{visMin}{distance}   = $2 + 0;
                $metar{visMin}{unitLength} = 'M';
                $metar{visMin}{compassDir} = $3;
            }
            $msg = $4;
        # EXTENSION: allow invalid formats
        } elsif ($msg =~ m@^(${re_vis_m}U) (.*)@o) {
            $metar{visPrev}{invalidFormat} = $1;
            $metar{visPrev}{s} = $1;
            $msg = $2;
        } else {
            # EXTENSION: allow missing visibility even without CAVOK
            $metar{WARNING} .= "WARNING: No visibility\n";
        }

=pod

After that may follow one or more runway visibilty ranges or C<RVRNO> if they
are not available.

=cut

        while ($msg =~ m@^(R($re_rwy_des)/(?:(////)|($re_rwy_vis)(?:V($re_rwy_vis))?(FT)?)(?:/)?([UDN])?) (.*)@o)
        {
            my $v;

            $v->{s}        = $1;
            $v->{rwyDesig} = $2;
            $v->{visTrend} = $7 if defined $7;
            if ($3) {
                $v->{RVR}{notAvailable} = 1;
            } else {
                $v->{RVR}{distance}   = $4;
                $v->{RVR}{unitLength} = defined $6 ? $6 : 'M';
                if (defined $5) {
                    $v->{RVRVariations}{distance}   = $5;
                    $v->{RVRVariations}{unitLength} = defined $6 ? $6 : 'M';
                    $v->{RVRVariations}{isLess}    = 1
                        if $v->{RVRVariations}{distance} =~ s/^M//;
                    $v->{RVRVariations}{isEqualGreater} = 1
                        if $v->{RVRVariations}{distance} =~ s/^P//;
                }
                $v->{RVR}{isLess}    = 1
                    if $v->{RVR}{distance} =~ s/^M//;
                $v->{RVR}{isEqualGreater} = 1
                    if $v->{RVR}{distance} =~ s/^P//;
            }
            push @{$metar{visRwy}}, $v;
            $msg = $8;
        }

        # EXTENSION: allow RVRNO (not available) (KADW, RKSG, PAED)
        while ($msg =~ m@^RVRNO (.*)@o) {
            $metar{RVRNO} = 1;
            $msg = $1;
        }

=pod

Then there can be up to 3 groups to describe the present weather.

=cut

        # EXTENSION: allow // (not available)
        while ($msg =~ m@^($re_weather|//|([+-]?(?:RATS|BRHZ|RESH|RETS|VC|VCRA|RABR))) (.*)@o) {
            if (defined $2) {
                push @{$metar{weather}}, { s => $2, invalidFormat => $2 };
            } else {
                push @{$metar{weather}}, _parseWeather $1;
            }
            $msg = $3;
        }

=pod

Then follow up to 3 groups for the sky condition (cloud cover and base or
vertical visibilty) optionally with cloud type. The keywords C<CLR>, C<SKC>, or
C<NSC> may indicate different sky conditions if no cloud cover is given.

=cut

        # EXTENSION: allow ////// (not available) without CB, TCU
        # WMO 306 Vol II
        #   CLR: no clouds below 10000 (FMH-1: 12000) ft detected by autom. st.
        # Annex 3 Appendix 3 4.5.4.1.e, f
        #   SKC: no clouds + VV not restricted but not CAVOK
        #   NSC: no significant clouds + no CB + VV not restr. but not CAVOK,SKC
        # AMOSSG/5-SoD Appendix C 4.9.1.4.b
        #   NCD: no clouds + CB, TCU detected by automatic observation system
        while ($msg =~ m@^((SKC|NSC|CLR|NCD)|VV($re_cloud_base)|(/{6}|$re_cloud_cov(?:$re_cloud_base$re_cloud_type?|$re_cloud_type))|($re_cloud_cov[0-9]{1,2}|$re_cloud_cov(?: [0-9]{1,3})?)) (.*)@o)
        {
            $msg = $6;
            if (defined $5) {
                push @{$metar{cloud}}, { s => $1, invalidFormat => $1 };
            } elsif (defined $2) {
                push @{$metar{cloud}}, { noClouds => $2, s => $1 };
            } elsif (defined $3) {
                $metar{visVert}{s} = $1;
                if ($3 eq '///') {
                    $metar{visVert}{notAvailable} = 1;
                } else {
                    $metar{visVert}{distance}   = $3 + 0;
                    $metar{visVert}{unitHeight} = 'FL';
                }
            } else {
                push @{$metar{cloud}}, _parseCloud $4;
            }
        }
        _determineCeiling $metar{cloud} if exists $metar{cloud};
    }

=pod

The following group contains the current air temperature and optionally the dew
point. If both are given the humidity can be determined.

=cut

    # EXTENSION: FMH-1: dew point is optional
    # EXTENSION: allow // for temperature and dew point
    # EXTENSION: allow XX for temperature and dew point
    if ($msg =~ '^((?:(M)?([0-9]{2})|(?://|XX))/((M)?([0-9]{2})|(?://|XX))?) (.*)')
    {
        $msg = $7;
        $metar{temperature}{s} = $1;
        if (defined $3) {
            $metar{temperature}{air}{temp} = $3 + 0;
            $metar{temperature}{air}{temp} *= -1 if defined $2;
            $metar{temperature}{air}{tempUnit} = 'C';
        } else {
            $metar{temperature}{air}{notAvailable} = 1;
        }
        if (defined $4) {
            if (defined $6) {
                $metar{temperature}{dewpoint}{temp} = $6 + 0;
                $metar{temperature}{dewpoint}{temp} *= -1 if defined $5;
                $metar{temperature}{dewpoint}{tempUnit} = 'C';
            } else {
                $metar{temperature}{dewpoint}{notAvailable} = 1;
            }
        }
        if (   exists $metar{temperature}{air}
            && exists $metar{temperature}{air}{temp}
            && exists $metar{temperature}{dewpoint}
            && exists $metar{temperature}{dewpoint}{temp})
        {
            my $t = $metar{temperature}{air}{temp};
            my $d = $metar{temperature}{dewpoint}{temp};

            # FGFS metar
            $metar{temperature}{relHumid1} =
                    100 * 10 ** (7.5 * ($d / ($d + 237.7) - $t / ($t + 237.7)));

            # http://www.bragg.army.mil/www-wx/wxcalc.htm
            # http://www.srh.noaa.gov/bmx/tables/rh.html
            $metar{temperature}{relHumid2} =
                      100 * ((112 - (0.1 * $t) + $d) / (112 + (0.9 * $t))) ** 8;

            # http://www.mattsscripts.co.uk/mweather.htm
            # http://ingrid.ldeo.columbia.edu/dochelp/QA/Basic/dewpoint.html
            $metar{temperature}{relHumid3} =
           100 * (6.11 * exp(5417.118093 * (1 / 273.15 - (1 / ($d + 273.15)))))
               / (6.11 * exp(5417.118093 * (1 / 273.15 - (1 / ($t + 273.15)))));

            # http://de.wikipedia.org/wiki/Taupunkt
            $metar{temperature}{relHumid4} =
                             100 * (6.11213 * exp(17.5043 * $d / (241.2 + $d)))
                                 / (6.11213 * exp(17.5043 * $t / (241.2 + $t)));
        }
    } else {
        $metar{WARNING} .= "WARNING: No temperature\n" unless $is_taf;
    }

=pod

The last regular item is the QNH given in hectopascal or in. Hg.

=cut

    # EXTENSION: allow // for last 2 digits
    # EXTENSION: allow tenths of hPa as .[0-9] (FLLS, HUEN, OPRN)
    # EXTENSION: allow 2 QNHs (OIxx), use first
    $ii = 0;
    while (   $ii < 2
        && $msg =~ '^((?:Q[01]|A[23])[0-9](?:[0-9]{2}|//)|[AQ]////|Q[01][0-9]{3}\.[[0-9]) (.*)')
    {
        $msg = $2;
        next if $ii++;

        $metar{QNH}{s} = $1;
        my ($unit, $dig12, $dig34) = $1 =~ '(.)(..)(.*)';
        if ("$dig12$dig34" eq '////') {
            $metar{QNH}{notAvailable} = 1;
        } else {
            $dig34 = '00' if $dig34 eq '//';
            if ($unit eq 'Q') {
                $metar{QNH}{hPa} = "$dig12$dig34" + 0;
            } else {
                $metar{QNH}{inHg} = "$dig12$dig34" / 100;
            }
        }
    }
    $metar{WARNING} .= "WARNING: No QNH\n"
        unless $is_taf || exists $metar{QNH};

=pod

After that may follow country or station specific information: pressure, worst
cloud cover. There may also be groups indicating recent weather, wind shear,
runway conditions and runway winds.

=cut

    # EXTENSION: allow some (sea level?) pressure (OIxx)
    if ($msg =~ m@^([789][0-9]{2}[./][0-9]) (.*)@o) {
        $msg = $2;
        $metar{somePressure}{s} = $1;
        $metar{somePressure}{hPa} = $1;
        $metar{somePressure}{hPa} =~ s@/@.@;
    }

    # EXTENSION: allow worst cloud cover
    if (   _isValidCountry('cloudMaxCover', \@cy)
        && $msg =~ m@^($re_cloud_cov|SKC) (.*)@o)
    {
        $msg = $2;
        $metar{cloudMaxCover}{s} = $1;
        if ($1 eq 'SKC') {
            $metar{cloudMaxCover}{SKC} = 1;
        } else {
            $metar{cloudMaxCover}{cloudCover} = $1;
        }
    }

    while ($msg =~ m@^RE($re_weather_re) (.*)@o) {
        $msg = $2;
        push @{$metar{recWeather}}, _parseWeather $1, 1;
    }

    if ($msg =~ '^(WS ALL RWY) (.*)') {
        $msg = $2;
        push @{$metar{windShear}}, { s => $1, rwyDesigAll => 1 };
    }
    while ($msg =~ m@^(WS RWY($re_rwy_des)) (.*)@o) {
        $msg = $3;
        push @{$metar{windShear}}, { s => $1, rwyDesig => $2 };
    }

    while (_parseRwyState \%metar, \$msg) {};

    # EXTENSION: allow colour code
    if ($msg =~ m@^($re_colour) (.*)@o) {
        $msg = $2;
        $metar{colourCode} = _parseColourCode $1;
    }

    # EXTENSION: country specific things
    while (   _isValidCountry('NEFO_PLAYA', \@cy)
           && $msg =~ m@^NEFO PLAYA ([1-9][0-9]+0FT|SKC) (.*)@o)
    {
        $msg = $2;
        $metar{NEFO_PLAYA} .= $1;
    }

    # EXTENSION:
    while (   _isValidCountry('rwyWind', \@cy)
           && $msg =~ m@^($re_rwy_wind) (.*)@o)
    {
        my $r;
        $msg = $6;
        $r->{s} = $1;
        $r->{wind} = _parseWind $3;
        $r->{wind}{windVarLeft} = $4 + 0 if defined $4;
        $r->{wind}{windVarRight} = $5 + 0 if defined $5;
        $r->{rwyDesig} = $2;
        push @{$metar{rwyWind}}, $r;
    }

=pod

Finally, there may be trends and remarks.

=cut

    if ($msg =~ '^(NOSIG) (.*)') {
        my $td;

        $td->{s} = $1;
        $td->{trendType} = $1;
        push @{$metar{trend}}, $td;
        $msg = $2;

        # EXTENSION: allow rwyState after NOSIG. Not considered trend data!
        while (_parseRwyState \%metar, \$msg) {};
    }

    # "supplementary" section of TAFs
    while ($is_taf && _parseTAFsuppl \%metar, \$msg, \%metar) {};

    # changes at midnight: 0000 with FM/AT, 2400 with TL
    # EXTENSION: allow missing period
    # EXTENSION: FM: allow time with Z
    # EXTENSION: FM: allow time without minutes
    while (   (!$is_taf && $msg =~ '^(BECMG|TEMPO) (.*)')
           || ($is_taf && $msg =~ m@^(FM($re_hour)($re_min)?Z?|(?:(BECMG|TEMPO)|(?:PROB([34]0)( TEMPO)?))(?: ($re_hour)($re_hour|24))?) (.*)@o))
   {
        my $td;

        if ($is_taf) {
            $msg = $9;
            $td->{s} = $1;
            if (defined $2) {
                $td->{trendType} = 'BECMG';
                @{$td->{trendTime1}}{'timeSpec', 'hour'} = ('FM', $2);
                $td->{trendTime1}{minute} = defined $3 ? $3 : '00';
            } else {
                if (defined $4) {
                    $td->{trendType} = $4;
                } else {
                    $td->{probability} = $5;
                    $td->{trendType} = defined $6 ? 'TEMPO' : 'PROB';
                }
                if (defined $7) {
                    @{$td->{trendTime1}}{'timeSpec', 'hour', 'minute'}
                                                             = ('FM', $7, '00');
                    @{$td->{trendTime2}}{'timeSpec', 'hour', 'minute'}
                                                             = ('TL', $8, '00');
                } else {
                    $metar{WARNING} .= "WARNING: period missing for $1\n";
                }
            }
        } else {
            $msg = $2;
            $td->{s} = $1;
            $td->{trendType} = $1;

            if ($msg =~ m@^((?:TL2400|(?:FM|TL|AT)$re_hour$re_min)Z?) (.*)@o) {
                $msg = $2;
                $td->{s} .= " $1";
                @{$td->{trendTime1}}{'timeSpec', 'hour', 'minute'} =
                                                           $1 =~ /(..)(..)(..)/;

                if ($msg =~ m@^TL(2400Z?|$re_hour${re_min}Z?) (.*)@o) {
                    $td->{s} .= " TL$1";
                    $msg = $2;
                    $td->{trendTime2}{timeSpec} = 'TL';
                    @{$td->{trendTime2}}{'hour', 'minute'} = $1 =~ /(..)(..)/;
                }
            }
        }

        if ($msg =~ m@^($re_wind) (.*)@o)
        {
            $msg = $6;
            $td->{sfcWind}{s} = $1;
            $td->{sfcWind}{wind} = _parseWind $1;
        }

        if ($msg =~ '^CAVOK (.*)') {
            $td->{CAVOK} = 1;
            $msg = $1;
        }

        if (!$td->{CAVOK}) {
            # EXTENSION: allow 'M' after re_vis_m
            if ($msg =~ m@^(($re_vis_m)M?|($re_vis_km)|(${re_vis_sm})SM) (.*)@o)
            {
                $msg = $5;
                $td->{visPrev}{s} = $1;
                if (defined $2) {
                    $td->{visPrev}{distance}   = $2 + 0;
                    $td->{visPrev}{unitLength} = 'M';
                } elsif (defined $3) {
                    $td->{visPrev}{distance}   = $3;
                    $td->{visPrev}{distance}   =~ s/KM//;
                    $td->{visPrev}{unitLength} = 'KM';
                } else {
                    my ($is_less, $vis) = $4 =~ '(M)?(.*)';
                    $td->{visPrev}{distance}   = _parseSM $vis;
                    $td->{visPrev}{isLess}     = 1 if defined $is_less;
                    $td->{visPrev}{unitLength} = 'SM';
                }
            } elsif ($is_taf && $msg =~ m@^P([1-9][0-9]*)SM (.*)@o) {
                $msg = $2;
                $td->{visPrev}{s} = "P$1SM";
                $td->{visPrev}{distance}       = $1;
                $td->{visPrev}{unitLength}     = 'SM';
                $td->{visPrev}{isEqualGreater} = 1;
            } elsif ($msg =~ m@^([0-9]{3}) (.*)@o) {
                $td->{visPrev}{s} = $1;
                $td->{visPrev}{invalidFormat} = $1;
                $msg = $2;
            }

            # EXTENSION: allow VC..
            while ($msg =~ m@^($re_weather|NSW) (.*)@o) {
                push @{$td->{weather}}, _parseWeather $1;
                $msg = $2;
            }

            $ii = 0;
            while ($msg =~ m@^((SKC|NSC)|VV($re_cloud_base)|($re_cloud_cov$re_cloud_base$re_cloud_type?|$re_cloud_cov$re_cloud_type)) (.*)@o)
            {
                $msg = $5;
                if (defined $2) {
                    push @{$td->{cloud}}, { noClouds => $2, s => $1 };
                } elsif (defined $3) {
                    $td->{visVert}{s} = $1;
                    if ($3 eq '///') {
                        $td->{visVert}{notAvailable} = 1;
                    } else {
                        $td->{visVert}{distance}   = $3 + 0;
                        $td->{visVert}{unitHeight} = 'FL';
                    }
                } else {
                    push @{$td->{cloud}}, _parseCloud $4;
                }
            }
            _determineCeiling $td->{cloud} if exists $td->{cloud};

            # EXTENSION: allow colour code
            if ($msg =~ m@^($re_colour) (.*)@o) {
                $msg = $2;
                $td->{colourCode} = _parseColourCode $1;
            }
        }

        # EXTENSION: allow rwyState after trend BECMG / TEMPO
        while (_parseRwyState $td, \$msg) {};

        # "supplementary" section of TAFs
        while ($is_taf && _parseTAFsuppl $td, \$msg, \%metar) {};

        push @{$metar{trend}}, $td;
    }

    if ($is_taf && $msg =~ m@^(TX(M)?([0-9]{2})/($re_hour|24)Z) (TN(M)?([0-9]{2})/($re_hour|24)Z) (.*)@o) {
        $msg = $9;
        $metar{tempMaxMin} = 1;
        $metar{tempMax}{s} = $1;
        $metar{tempMax}{temp} = $3 + 0;
        $metar{tempMax}{temp} *= -1 if defined $2;
        $metar{tempMax}{tempUnit} = 'C';
        $metar{tempMax}{hour} = $4;
        $metar{tempMin}{s} = $5;
        $metar{tempMin}{temp} = $7 + 0;
        $metar{tempMin}{temp} *= -1 if defined $6;
        $metar{tempMin}{tempUnit} = 'C';
        $metar{tempMin}{hour} = $8;
    }

    if ($is_taf && $msg =~ m@^(TN(M)?([0-9]{2})/($re_hour|24)Z) (TX(M)?([0-9]{2})/($re_hour|24)Z) (.*)@o) {
        $msg = $9;
        $metar{tempMin}{s} = $1;
        $metar{tempMin}{temp} = $3 + 0;
        $metar{tempMin}{temp} *= -1 if defined $2;
        $metar{tempMin}{tempUnit} = 'C';
        $metar{tempMin}{hour} = $4;
        $metar{tempMax}{s} = $5;
        $metar{tempMax}{temp} = $7 + 0;
        $metar{tempMax}{temp} *= -1 if defined $6;
        $metar{tempMax}{tempUnit} = 'C';
        $metar{tempMax}{hour} = $8;
    }

    if ($msg =~ '^RMK (.*)') {
        my ($notRecognized, $parsed);

        @{$metar{remark}} = ();
        $msg = $1;

        $notRecognized = '';
        while ($msg ne '') {
            my ($r, $rc, $ft, %wind);

            $parsed = 1;
            $r = {};
            if (_parseRwyState $r, \$msg) {
                push @{$metar{remark}}, { rwyState => $r->{rwyState}[0] };
            } elsif (   _isValidCountry('remarkCloudMaxCov', \@cy)
                && $msg =~ m@^($re_cloud_cov) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, {
                    cloudMaxCover => { s => $1, cloudCover => $1 },
                };
            } elsif (   _isValidCountry('remarkColour', \@cy)
                && $msg =~ m@^($re_colour) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { colourCode => _parseColourCode $1 };
            } elsif (   _isValidCountry('remarkCloudType', \@cy)
                     && $msg =~ m@^(8/([0-9/])([0-9/])([0-9/])) (.*)@o)
            {
                $msg = $5;
                $r->{s} = $1;
                if ($2 eq '/') {
                    $r->{cloudTypeLowAboveOvercast} = 1;
                } else {
                    $r->{cloudTypeLow} = $2;
                }
                if ($3 eq '/') {
                    $r->{cloudTypeMiddleAboveOvercast} = 1;
                } else {
                    $r->{cloudTypeMiddle} = $3;
                }
                if ($4 eq '/') {
                    $r->{cloudTypeHighAboveOvercast} = 1;
                } else {
                    $r->{cloudTypeHigh} = $4;
                }
                push @{$metar{remark}}, { cloudTypeFamily => $r };
            } elsif ($msg =~ m@^(PWINO|FZRANO|TSNO|PNO|RVRNO|PRES[FR]R|AURBO|NOSPECI|RCRNR) (.*)@o){
                $msg = $2;
                push @{$metar{remark}}, { $1 => 1 };
            } elsif ($msg =~ m@^(?:\$) (.*)@o){
                $msg = $1;
                push @{$metar{remark}}, { needMaint => { s => '$' }};
            } elsif ($msg =~ m@^((?:(LGT) )?(FROIN)) (.*)@o){
                $msg = $4;
                $ft->{s}   = $1;
                $ft->{txt} = ($2 ? $keywords{$2} . ' ' : '') . $keywords{$3};
            } elsif ($msg =~ m@^(SD[FGP]/HD[FGP]) (.*)@o){
                $msg = $2;
                $ft->{s}   = $1;
                $ft->{txt} = "??:\t\t\t$1";
            } elsif (   _isValidCountry('remarkCIGRAG', \@cy)
                     && $msg =~ m@^(CIG (?:RAG|RGD)) (.*)@o)
            {
                $msg = $2;
                $ft->{s}   = $1;
                $ft->{txt} = $keywords{CIGRAG};
            } elsif (   _isValidCountry('remarkCIGVRBL', \@cy)
                     && $msg =~ m@^(CIG VRBL? ($re_vis_sm)-($re_vis_sm)) (.*)@o)
            {
                $msg = $4;
                $ft->{s}   = $1;
                $ft->{txt} = 'ceiling is 900 meters or less and variable, ' .
                             "visibility varies from $2 to $3 SM";
            } elsif (   _isValidCountry('remarkSLPdPa', \@cy)
                     && $msg =~ m@^(SLP(?:([0-9]{2})([0-9])|NO)) (.*)@o)
            {
                $msg = $4;

                if (!defined $2) {
                    push @{$metar{remark}}, { SLPdPa => {
                        s => $1,
                        notAvailable => 1
                    }};
                # TODO: really need QNH? but values are REALLY close:
                # KLWM 161454Z AUTO 31012G21KT 10SM CLR M10/M18 A2808 RMK SLP511
                #   -> 951.1 hPa
                # KQHT 021355Z 26003KT 8000 IC M15/M16 A3070 RMK SLP484
                #   -> 1048.4 hPa
                # higher QNHs possible:
                # ETNS 161520Z 11006KT 7000 FEW020 M02/M05 Q1078 WHT WHT
                #   -> SLP78?
                # UNWW 021400Z 00000MPS CAVOK M25/M29 Q1056 NOSIG RMK QFE759
                #   -> SLP56?
                } elsif (!exists $metar{QNH} || !exists $metar{QNH}{inHg}) {
                    if ($2 > 80 || $2 < 20) { # only if within sensible range
                        my $slp;
                        $slp = "$2.$3";
                        $slp += $slp < 50 ? 1000 : 900;
                        push @{$metar{remark}}, { SLPdPa => {
                            s => $1,
                            hPa => sprintf '%.1f', $slp
                        }};
                    } else {
                        push @{$metar{remark}}, { SLPdPa => {
                            s => $1,
                            invalidFormat => "no QNH, x$2.$3 hPa"
                        }};
                    }
                } else {
                    my ($slp, @slp, @qnh, $qnhHPa);

                    @slp = ($1, $2, $3);
                    $qnhHPa = rnd($metar{QNH}{inHg} * $INHG2HPA, 1);

                    $slp = $qnhHPa;                       # start with given QNH
                    $slp =~ s/..$//;                  # remove 2 trailing digits
                    $slp .= "$slp[1].$slp[2]";              # append given value
                    $slp += 100 if $slp + 50 < $qnhHPa;     # ensure SLP > QNH
                    push @{$metar{remark}}, { SLPdPa => {
                        s => $slp[0],
                        hPa => sprintf '%.1f', $slp
                    }};
                }
            } elsif (   _isValidCountry('remarkSLPinHg', \@cy)
                     && $msg =~ m@^(SLP(?:([0-9]{3})|NO)) (.*)@o)
            {
                $msg = $3;

                if (!defined $2) {
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $1,
                        notAvailable => 1
                    }};
                } elsif (!exists $metar{QNH} || !exists $metar{QNH}{inHg}) {
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $1,
                        invalidFormat => "no QNH, x$2 in. Hg"
                    }};
                } else {
                    my ($qnhInHg, @slp);

                    @slp = ($1, $2);
                    $qnhInHg = $metar{QNH}{inHg} * 100;
                    $qnhInHg =~ s/...$//;
                    $qnhInHg .= "$slp[1]";
                    $qnhInHg /= 100;
                    $qnhInHg += 10 if $qnhInHg + 5 < $metar{QNH}{inHg};
                    push @{$metar{remark}}, { SLPinHg => {
                        s => $slp[0],
                        inHg => $qnhInHg
                    }};
                }
            } elsif (   _isValidCountry('remarkGRIDWind', \@cy)
                     && $msg =~ m@^GRID(${re_wind_dir}0${re_wind_speed}KT) (.*)@o)
            {
                $msg = $2;
                $r->{gridWind}{s} = "GRID$1";
                $r->{gridWind}{wind} = _parseWind $1;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkRwyWind', \@cy)
                     && $msg =~ m@^($re_rwy_wind) (.*)@o)
            {
                $msg = $6;
                $r->{rwyWind}{s} = $1;
                $r->{rwyWind}{rwyDesig} = $2;
                $r->{rwyWind}{wind} = _parseWind $3;
                $r->{rwyWind}{wind}{windVarLeft} = $4 + 0 if defined $4;
                $r->{rwyWind}{wind}{windVarRight} = $5 + 0 if defined $5;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkRsc', \@cy)
                     && $msg =~ m@^($re_rsc) (.*)@o)
            {
                $msg = $2;
                $delim = '';
                $ft->{s}   = $1;
                $ft->{txt} = sprintf "rwy surface condition:\t";
                for ($1 =~ /(SLR|LSR|PSR|P|SANDED|WET|DRY|IR|\/\/|[0-9]{2})+?/g) {
                    $ft->{txt} .=  $delim;
                    if ($_ =~ '[0-9]') {
                        $ft->{txt} .=  "decelerometer reading $_";
                    } else {
                        $ft->{txt} .= $_ eq '//'
                                        ? 'no decelerometer reading available'
                                        : $keywords{$_};
                    }
                    $delim = ', ';
                }
            } elsif (   _isValidCountry('remarkCloudOpacity', \@cy)
                     && $msg =~ m@^(((?:$re_opac_cloud[0-8])+)(?: ($re_cloud_type) ASOCTD)?) (.*)@o)
            {
                $ft->{s} = $1;
                $ft->{txt} = _printCldWOpacity $2;
                $ft->{txt} .= "\n\t\t" . $cloud_types{$3} .
                              ' ' . $keywords{ASOCTD}
                    if defined $3;
                $msg = $4;
            } elsif (   _isValidCountry('remarkCloudOpacityRev', \@cy)
                     && $msg =~ m@^([0-8])($re_opac_cloud) (.*)@o)
            {
                $ft->{s}   = "$1$2";
                $ft->{txt} = _printCldWOpacity "$2$1";
                $msg = $3;
            } elsif (   _isValidCountry('remarkCloudOpacityLvl', \@cy)
                     && $msg =~ m@^(([0-8])(?:((?:BL)?SN|FG)|($re_cloud_type))([0-9]{3})($re_phen_locdirs)?) (.*)@o)
            {
                $msg = $7;
                $r->{s} = $1;
                $r->{eights} = $2;
                $r->{weather} = _parseWeather $3 if defined $3;
                $r->{cloudType} = $4 if defined $4;
                $r->{cloudBase} = $5 + 0;
                $r->{locationAndList} = _parseLocations $6 if defined $6;
                push @{$metar{remark}}, { cloudOpacityLvl => $r };
            } elsif (   _isValidCountry('remarkCloudCoverVar', \@cy)
                     && $msg =~ m@^(($re_cloud_cov$re_cloud_base?) V ($re_cloud_cov)) (.*)@o)
            {
                $msg = $4;
                $r->{cloudCoverVar} = _parseCloud $2;
                $r->{cloudCoverVar}{s} = $1;
                $r->{cloudCoverVar}{cloudCover2} = $3;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkCloudCoverLvl', \@cy)
                     && $msg =~ m@^($re_cloud_cov$re_cloud_base) (.*)@o)
            {
                $msg = $2;
                $r->{cloud} = _parseCloud $1;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkCloudTypeLvl', \@cy)
                     && $msg =~ m@^($re_cloud_type)($re_cloud_base) (.*)@o)
            {
                $msg = $3;
                $ft->{s}   = "$1$2";
                $ft->{txt} = $cloud_types{$1} . ' at ' . $2 * 100 . ' ft';
            } elsif (   _isValidCountry('remarkReWeather', \@cy)
                     && $msg =~ m@^RE($re_weather_re) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { recWeather => [ _parseWeather($1) ] };
            } elsif (   _isValidCountry('remarkWR', \@cy)
                     && $msg =~ m@^(WR(//)?) (.*)@o)
            {
                $msg = $3;
                $r->{wetRunway}{s}        = $1;
                $r->{wetRunway}{BANotRep} = 1 if defined $2;
                $r->{freetext}{s}   = $1;
                $r->{freetext}{txt} = "Runway condition:\twet runway";
                $r->{freetext}{txt} .= ", braking action not reported"
                    if defined $2;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkSfcVisM', \@cy)
                     && $msg =~ m@^SFC VIS (($re_vis_m)|($re_vis_km)) (.*)@o)
            {
                $msg = $4;
                $r->{s} = "SFC VIS $1";
                $r->{locationAt} = 'SFC';
                if (defined $2) {
                    $r->{visibility}{distance}   = $2 + 0;
                    $r->{visibility}{unitLength} = 'M';
                } else {
                    $r->{visibility}{distance}   = $3;
                    $r->{visibility}{distance}   =~ s/KM//;
                    $r->{visibility}{unitLength} = 'KM';
                }
                push @{$metar{remark}}, { visAt => $r };
            } elsif (   _isValidCountry('remarkVisSfcTwr', \@cy)
                     && $msg =~ m@^(SFC|TWR) VIS ($re_vis_sm) (.*)@o)
            {
                $msg = $3;
                $r->{s} = "$1 VIS $2";
                $r->{locationAt} = $1;
                my ($is_less, $vis) = $2 =~ '(M)?(.*)';
                $r->{visibility}{distance}   = _parseSM $vis;
                $r->{visibility}{unitLength} = 'SM';
                $r->{visibility}{isLess}     = 1 if $is_less;
                push @{$metar{remark}}, { visAt => $r };
            } elsif (   _isValidCountry('remarkVisVarK', \@cy)
                     && $msg =~ m@^VIS (${re_vis_sm}V$re_vis_sm) (.*)@o)
            {
                $msg = $2;
                $r->{visVar1}{s} = "VIS $1";
                my ($is_less1, $vis1, $is_less2, $vis2) =
                                               $1 =~ '(M)?([^V]+)V(M)?(.*)';
                $r->{visVar1}{distance}   = _parseSM $vis1;
                $r->{visVar1}{unitLength} = 'SM';
                $r->{visVar1}{isLess}     = 1 if $is_less1;
                $r->{visVar2}{distance}   = _parseSM $vis2;
                $r->{visVar2}{unitLength} = 'SM';
                $r->{visVar2}{isLess}   = 1 if $is_less2;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkVisVarCY', \@cy)
                     && $msg =~ m@^VIS VRB ($re_vis_sm-$re_vis_sm) (.*)@o)
            {
                $msg = $2;
                $r->{visVar1}{s} = "VIS VRB $1";
                my ($is_less1, $vis1, $is_less2, $vis2) =
                                               $1 =~ '(M)?([^-]+)-(M)?(.*)';
                $r->{visVar1}{distance} = _parseSM $vis1;
                $r->{visVar1}{unitLength} = 'SM';
                $r->{visVar1}{isLess} = 1 if $is_less1;
                $r->{visVar2}{distance} = _parseSM $vis2;
                $r->{visVar2}{unitLength} = 'SM';
                $r->{visVar2}{isLess} = 1 if $is_less2;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkVisVarCY', \@cy)
                     && $msg =~ m@^VSBY VRBL ([0-9]\.[0-9]V[0-9]\.[0-9]) (.*)@o)
            {
                $msg = $2;
                $r->{visVar1}{s} = "VSBY VRBL $1";
                my ($is_less1, $vis1, $is_less2, $vis2) =
                                               $1 =~ '(M)?([^V]+)V(M)?(.*)';
                $r->{visVar1}{distance} = _parseSM $vis1;
                $r->{visVar1}{unitLength} = 'SM';
                $r->{visVar1}{isLess} = 1 if $is_less1;
                $r->{visVar2}{distance} = _parseSM $vis2;
                $r->{visVar2}{unitLength} = 'SM';
                $r->{visVar2}{isLess} = 1 if $is_less2;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkPCPNLastHour', \@cy)
                     && $msg =~ m@^(PCPN ([0-9]+\.[0-9])MM (PAST HR)) (.*)@o)
            {
                $msg = $4;
                $ft->{s}   = $1;
                $ft->{txt} = "@keywords{'PCPN', $3}: $2 mm";
            } elsif (   _isValidCountry('remarkVisLocM', \@cy)
                     && $msg =~ m@^VIS((?:$re_phen_locdirs $re_vis_m_km_remark)(?:(?: AND)?$re_phen_locdirs $re_vis_m_km_remark)*) (.*)@o)
            {
                my @arr;

                $msg = $2;
                $r->{visListLoc}{s} = "VIS$1";
                $r->{visListLoc}{arr} = \@arr;
                for ($1 =~ m@(?:AND)?($re_phen_locdirs $re_vis_m_km_remark)@og) {
                    my $v;

                    m@($re_phen_locdirs) (?:($re_vis_m)|([1-9][0-9]{2,3})M|($re_vis_km))@o;
                    $v->{locationAndList} = _parseLocations $1;
                    if (defined $2) {
                        $v->{visibility}{distance}   = $2 + 0;
                        $v->{visibility}{unitLength} = 'M';
                    } elsif (defined $3) {
                        $v->{visibility}{distance}   = $3  + 0;
                        $v->{visibility}{unitLength} = 'M';
                    } else {
                        $v->{visibility}{distance} = $4;
                        $v->{visibility}{distance} =~ s/KM//;
                        $v->{visibility}{unitLength} = 'KM';
                    }
                    push @arr, $v;
                }
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkVisLwr', \@cy)
                     && $msg =~ m@^(VIS LWR($re_phen_locdirs)) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { visLower => {
                    s => $1,
                    locationAndList => _parseLocations $2
                }};
            } elsif (   _isValidCountry('remarkVisLwr', \@cy)
                     && $msg =~ m@^(VIS($re_phen_locdirs) LWR) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { visLower => {
                    s => $1,
                    locationAndList => _parseLocations $2
                }};
            } elsif (   _isValidCountry('remarkCigLwr', \@cy)
                     && $msg =~ m@^(CIG LWR($re_phen_locdirs)) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { cigLower => {
                    s => $1,
                    locationAndList => _parseLocations $2
                }};
            } elsif (   _isValidCountry('remarkQNH', \@cy)
                     && $msg =~ m@^(A)([23][0-9]{3}) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { QNH => {
                    s => "$1$2",
                    inHg => $2 / 100
                }};
            } elsif (   _isValidCountry('remarkQFF', \@cy)
                     && $msg =~ m@^([789][0-9]{2}) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { QFF => {
                    s => $1,
                    hPa => $1
                }};
            } elsif (   _isValidCountry('remarkCorrected', \@cy)
                     && $msg =~ m@^(COR ($re_hour)($re_min)) (.*)@o)
            {
                $msg = $4;
                $r->{correctedAt}{s} = $1;
                $r->{correctedAt}{hour} = $2;
                $r->{correctedAt}{minute} = $3;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkSnowIncr', \@cy)
                     && $msg =~ m@^(SNINCR ([0-9]+)/([0-9]+)) (.*)@o)
            {
                $msg = $4;
                $r->{snowIncr}{s} = $1;
                $r->{snowIncr}{pastHour} = $2;
                $r->{snowIncr}{onGround} = $3;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkHourly', \@cy)
                     && $msg =~ m@^1([01][0-9]{3}) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { temp6hMax => {
                    s        => "1$1",
                    temp     => _parseUSTemp($1),
                    tempUnit => 'C'
                }};
            } elsif (   _isValidCountry('remarkHourly', \@cy)
                     && $msg =~ m@^2([01][0-9]{3}) (.*)@o)
            {
                $msg = $2;
                push @{$metar{remark}}, { temp6hMin => {
                    s        => "2$1",
                    temp     => _parseUSTemp($1),
                    tempUnit => 'C'
                }};
            } elsif (   _isValidCountry('remarkHourly', \@cy)
                     && $msg =~ m@^(?:5([0-8][0-9]{3}||////)) (.*)@o)
            {
                $msg = $2;

                $ft->{s}   = "5$1";
                $1 =~ /(.)(.*)/;
                $ft->{txt} = "3-hourly press. tend.:\t"
                        . { 0 => 'increasing, then decreasing',
                            1 => 'increasing, then steady, or increasing then increasing more slowly',
                            2 => 'increasing steadily or unsteadily',
                            3 => 'decreasing or steady, then increasing; or increasing then increasing more rapidly',
                            4 => 'steady',
                            5 => 'decreasing then increasing',
                            6 => 'decreasing, then steady, or decreasing then decreasing more slowly',
                            7 => 'decreasing steadily or unsteadily',
                            8 => 'Steady or increasing, then decreasing; or decreasing then decreasing more rapidly',
                           '/' => '(not available)',
                        }->{$1};
                $ft->{txt} .= " by " . ($2 / 10) . " hPa" unless $2 eq '///';
            } elsif (   _isValidCountry('remarkPeakWind', \@cy)
                     && $msg =~ m@^(PK WND (${re_wind_dir}0$re_wind_speed)/(?:($re_hour)?($re_min))) (.*)@o)
            {
                my ($wind, $time);

                $msg = $5;
                $r->{peakWind}{s} = $1;
                $r->{peakWind}{wind} = _parseWind "$2KT";
                $r->{peakWind}{hour} = $3 if defined $3;
                $r->{peakWind}{minute} = $4;
                push @{$metar{remark}}, $r;
            } elsif (   _isValidCountry('remarkHourlyTemp', \@cy)
                     && $msg =~ m@^(T([01][0-9]{3})([01][0-9]{3})?) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                $r->{air}{temp}          = _parseUSTemp($2);
                $r->{air}{tempUnit}      = 'C';
                if (defined $3) {
                    $r->{dewpoint}{temp}     = _parseUSTemp($3);
                    $r->{dewpoint}{tempUnit} = 'C';
                }
                push @{$metar{remark}}, { tempHourly => $r };
            } elsif (   _isValidCountry('remarkQFEhPa', \@cy)
                     && $msg =~ m@^(QFE([01]?[0-9]{3})) (.*)@o)
            {
                $msg = $3;
                push @{$metar{remark}}, { QFE => {
                    s => $1,
                    hPa => $2 + 0
                }};
            } elsif (   _isValidCountry('remarkDA_PA', \@cy)
                     && $msg =~ m@^((DA|PA)[ /]?([+-]?[0-9]+)(?:FT)?) (.*)@o)
            {
                $msg = $4;
                $ft->{s}   = $1;
                $ft->{txt} = $keywords{$2} . ":\t$3 ft";
            } elsif (   _isValidCountry('remarkKNIP', \@cy)
                     && $msg =~ m@^((RH|SST|AI|OAT)/([0-9]{2})F?) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                if ($2 eq 'SST' || $2 eq 'OAT') {
                    $r->{temp} = $3 + 0;
                    $r->{tempUnit} = 'F';
                } else {
                    $r->{"$2Val"} = $3 + 0;
                }
                push @{$metar{remark}}, { $2 => $r };
            } elsif (   _isValidCountry('remarkKN', \@cy)
                     && $msg =~ m@^((RH|SST)/([0-9]+)) (.*)@o)
            {
                $msg = $4;
                $r->{s} = $1;
                if ($2 eq 'SST') {
                    $r->{temp} = $3 + 0;
                    $r->{tempUnit} = 'C';
                } else {
                    $r->{"$2Val"} = $3 + 0;
                }
                push @{$metar{remark}}, { $2 => $r };
            } elsif (   _isValidCountry('remarkQFEmmHg', \@cy)
                     && $msg =~ m@^(QFE ?([0-9]{3})(?:/([0-9]{3,4}))?) (.*)@o)
            {
                my $qfe;
                $msg = $4;
                $qfe->{s} = $1;
                $qfe->{mmHg} = $2 + 0;
                $qfe->{hPa} = $3 + 0 if defined $3;
                push @{$metar{remark}}, { QFE => $qfe };
            } elsif ($msg =~ m@^(((?:VIS|CHI)NO) RWY ($re_rwy_des)) (.*)@o) {
                $msg = $4;
                $ft->{s}   = $1;
                $ft->{txt} = "@keywords{$2, 'RWY'} $3";
            } elsif (   _isValidCountry('remarkSNWCVR', \@cy)
                     && $msg =~ m@^(SNW? ?(?:CVR|COV)[/ ](MCH LOOSE|TRACE ?LOOSE|MED(?:IUM) PACK|HARD PACKED|TRLOOSE|NIL|TRACELOOSE)) (.*)@o)
            {
                $msg = $3;
                $ft->{s}   = $1;
                $ft->{txt} = "snow cover: $2";
            } elsif (   _isValidCountry('remarkObsTaken', \@cy)
                     && $msg =~ m@^(OBS TAKEN [+]([0-9]+)) (.*)@o)
            {
                $msg = $3;
                $ft->{s}   = $1;
                $ft->{txt} = "observation taken $2";
            } elsif (   _isValidCountry('remarkBalloon', \@cy)
                     && $msg =~ m@^((?:BLN (?:DSA?PRD ([0-9]+) ?FT[.]?|VI?SBL(?: TO)? ([0-9]+) ?FT)|(?:CIG )?BLN ESTM?D)) (.*)@o)
            {
                $msg = $4;
                $ft->{s} = $1;
                if (defined $2) {
                    $ft->{txt} = "balloon disappeared at $2 feet";
                } elsif (defined $3) {
                    $ft->{txt} = "balloon visible to $3 feet";
                } else {
                    $ft->{txt} = "ceiling from estimated balloon data";
                }
            } elsif (   _isValidCountry('remarkRVR', \@cy)
                     && $msg =~ m@^RVR((?: RWY ?$re_rwy_des $re_rwy_vis ?FT[.]?)+) (.*)@o)
            {
                my $hdr = 'RVR ';
                $msg = $2;
                for ($1 =~ / (RWY ?$re_rwy_des $re_rwy_vis ?FT[.]?)/g) {
                    my $v;

                    $v->{s} = "$hdr$_";
                    $hdr    = '';

                    /RWY ?($re_rwy_des) ($re_rwy_vis) ?(FT)/;

                    $v->{rwyDesig}        = $1;
                    $v->{RVR}{distance}   = $2;
                    $v->{RVR}{unitLength} = $3;
                    $v->{RVR}{isLess}    = 1
                        if $v->{RVR}{distance} =~ s/^M//;
                    $v->{RVR}{isEqualGreater} = 1
                        if $v->{RVR}{distance} =~ s/^P//;
                    push @{$metar{remark}}, { visRwy => $v };
                }
            } elsif (   _isValidCountry('remarkTAFNxtFcstBy', \@cy)
                     && $msg =~ m@^(NXT FCST BY ($re_hour)Z) (.*)@o)
            {
                $msg = $3;
                $ft->{s} = $1;
                $ft->{txt} = "next forecast by $2:00 UTC";
            } elsif (   _isValidCountry('remarkTAFNxtFcstAt', \@cy)
                     && $msg =~ m@^(NXT FCST WILL BE ISSUED AT ($re_day)($re_hour)($re_min)Z) (.*)@o)
            {
                $msg = $5;
                $ft->{s} = $1;
                $ft->{txt} = "next forecast on $2., at $3:$4 UTC";
            } elsif (   _isValidCountry('remarkTAFFcstAutoObs', \@cy)
                     && $msg =~ m@^(FCST BASED ON AUTO OBS[.]) (.*)@o)
            {
                $msg = $2;
                $ft->{s} = $1;
                $ft->{txt} = "forecast based on automatic observation";
            } elsif (   _isValidCountry('remarkTAFFcstAutometar', \@cy)
                     && $msg =~ m@^(BASED ON AUTOMETAR) (.*)@o)
            {
                $msg = $2;
                $ft->{s} = $1;
                $ft->{txt} = "forecast based on automatic METAR";
            } elsif (   _isValidCountry('remarkPhenCloud', \@cy)
                     && $msg =~ m@^((FU|(?:FZ)?FG|BR|PWR PLNT(?: PLUME)?) ($re_cloud_cov$re_cloud_base)) (.*)@o)
            {
                my ($cl, $phen);

                $msg = $4;
                $ft->{s} = $1;
                $phen = $2;
                $cl = _parseCloud $3;
                if ($phen =~ m@^$re_weather$@) {
                    $ft->{txt} = printWeather(_parseWeather $phen);
                } else {
                    $ft->{txt} = $keywords{$phen};
                }
                $ft->{txt} .= ' ' . $keywords{OBSCG} . ' ' .
                      { FEW => '1/8 - 2/8',
                        SCT => '3/8 - 4/8',
                        BKN => '5/8 - 7/8',
                        OVC => '8/8' }->{$cl->{cloudCover}}
                      . ' of the sky';
                if ($cl->{cloudBase} eq '000') {
                    $ft->{txt} .= ' at ground level';
                } else {
                    $ft->{txt} .= ' based at ' . $cl->{cloudBase} * 100 . ' ft'
                }

# nearly the same pattern twice except for the position of $re_phen_desc
# (do not match too much but don't miss anything, either)

            } elsif ($msg =~ m@^((?:($re_phen_desc) )(?:(LTG|VIRGA|FG BNK|PCPN|FULYR|HZY|BINOVC|ICG)|($re_cloud_type(?:[/-]$re_cloud_type)*)|($re_weather)|($re_cloud_cov))(?:($re_phen_locdirs)|(?: (DSNT))?($re_num_quadrant))?$re_wx_mov_d?) (.*)@o)
            {
                $msg = $12;
                $r->{s} = $1;
                _parsePhenomDescr $r, $2 if defined $2;
                if (defined $3) {
                    $r->{otherPhenom} = $3;
                } elsif (defined $4) {
                    @{$r->{cloudType}} = split m@[/-]@, $4;
                } elsif (defined $5) {
                    $r->{weather} = _parseWeather $5;
                } else {
                    $r->{cloudCover} = $6;
                }
                $r->{locationAndList} = _parseLocations $7 if defined $7;
                $r->{locationAndList} = _parseQuadrants $9, $8 if defined $9;
                $r->{$10} = _parseLocations $11 if defined $11;
                $r->{otherPhenom} =~ s/ /_/ if exists $r->{otherPhenom};
                push @{$metar{remark}}, { phenomLocation => $r };
            } elsif ($msg =~ m@^((?:(LTG|VIRGA|FG BNK|PCPN|FULYR|HZY|BINOVC|ICG)|($re_cloud_type(?:[/-]$re_cloud_type)*)|($re_weather)|($re_cloud_cov))(?: ($re_phen_desc))?(?:($re_phen_locdirs)|(?: (DSNT))?($re_num_quadrant))?$re_wx_mov_d?) (.*)@o)
            {
                $msg = $12;
                $r->{s} = $1;
                if (defined $2) {
                    $r->{otherPhenom} = $2;
                } elsif (defined $3) {
                    @{$r->{cloudType}} = split m@[/-]@, $3;
                } elsif (defined $4) {
                    $r->{weather} = _parseWeather $4;
                } else {
                    $r->{cloudCover} = $5;
                }
                _parsePhenomDescr $r, $6 if defined $6;
                $r->{locationAndList} = _parseLocations $7 if defined $7;
                $r->{locationAndList} = _parseQuadrants $9, $8 if defined $9;
                $r->{$10} = _parseLocations $11 if defined $11;
                $r->{otherPhenom} =~ s/ /_/ if exists $r->{otherPhenom};
                push @{$metar{remark}}, { phenomLocation => $r };
            } elsif ($msg =~ m@^(($re_phen_desc) LTG($re_ltg_types+)($re_phen_locdirs)?$re_wx_mov_d?) (.*)@o)
            {
                my $txt_lt;

                $msg = $7;
                $r->{s} = $1;
                _parsePhenomDescr $r, $2 if defined $2;
                $r->{locationAndList} = _parseLocations $4 if defined $4;
                $r->{$5} = _parseLocations $6 if defined $6;
                $r->{lightningType} = ();
                for ($3 =~ /$re_ltg_types/og) {
                    push @{$r->{lightningType}}, $_;
                }
                push @{$metar{remark}}, { phenomLocation => $r };
            } elsif ($msg =~ m@^(AO[12]) (.*)@o) {
                $msg = $2;
                push @{$metar{remark}}, { obsStationType => {
                    s           => $1,
                    stationType => $1
                }};
            } elsif ($msg =~ m@^(CIG ($re_cloud_base)(?:(?:(?: (APCH))?(?: RWY ?| R?)?($re_rwy_des))($re_phen_locdirs)?|($re_phen_locdirs))) (.*)@o)
            {
                $msg = $7;
                $ft->{s}   = $1;
                $ft->{txt} = $keywords{CIG} . ' ' . $2 * 100 . ' ft';
                $ft->{txt} .= ' ' . $keywords{$3} if defined $3;
                $ft->{txt} .= ' runway ' . $4 if defined $4;
                $ft->{txt} .= printLocations(_parseLocations $5) if defined $5;
                $ft->{txt} .= printLocations(_parseLocations $6) if defined $6;
            } elsif ($msg =~ m@^(NOSIG|CAVU|EPC|EPO|RTS) (.*)@o) {
                $msg = $2;
                $ft->{s}   = $1;
                $ft->{txt} = $keywords{$1};
            } elsif (   $metar{obsStation}{id} eq 'ZMUB'
                     && $msg =~ m@^([0-9] [0-9][0-9]) (.*)@o) {
                $msg = $2;
                $ft->{s}   = $1;
                $ft->{txt} = "??:\t\t\t$1";
            } elsif ($cy[1] eq 'LF' && $msg =~ m@^([MB])([0-9]) (.*)@o) {
                $msg = $3;
                $r->{reportConcerns}{s} = "$1$2";
                $r->{reportConcerns}{change} = $1;
                $r->{reportConcerns}{subject} = $2;
                push @{$metar{remark}}, $r;
            } elsif ($cy[1] eq 'LI') {
                my $re_cond_moun =
                    '(?:LIB|CLD SCT|VERS INC|CNS POST|CLD CIME'
                    . '|CIME INC|GEN INC|INC|INVIS)';
                my $re_chg_moun =
                    '(?:NC|CUF'
                     . '|ELEV (?:SLW|RAPID|STF)'
                     . '|ABB (?:SLW|RAPID)'
                     . '|STF(?: ABB)?'
                     . '|VAR RAPID)';
                my $re_cond_vall =
                    '(?:NIL|FOSCHIA(?: SKC SUP)?|NEBBIA(?: SCT)?'
                    . '|CLD SCT(?: NEBBIA INF)?|MAR CLD|INVIS)';
                my $re_chg_vall =
                    '(?:NC|DIM(?: ELEV| ABB)?|AUM(?: ELEV| ABB)?|ABB'
                    . '|NEBBIA INTER)';
                my $re_moun =
                            "$re_phen_locdirs? $re_cond_moun(?: $re_chg_moun)?";
                my $re_vall =
                            "$re_phen_locdirs? $re_cond_vall(?: $re_chg_vall)?";
                if ($msg =~ m@^(MON((?:$re_moun)+)) (.*)@o) {
                    $msg = $3;
                    $r->{s} = $1;
                    $r->{condMoun} = ();
                    for ($2 =~ m@($re_moun)@og) {
                        my $m;
                        m@($re_phen_locdirs)? ($re_cond_moun)(?: ($re_chg_moun))?@o;

                        $m->{locationAndList} =_parseLocations $1 if defined $1;
                        $m->{condMounType}    = $2;
                        $m->{condMounChange}  = $3 if defined $3;
                        $m->{condMounType} =~ s/ /_/g;
                        $m->{condMounChange} =~ s/ /_/g
                            if exists $m->{condMounChange};
                        push @{$r->{condMoun}}, $m;
                    }
                    push @{$metar{remark}}, { conditionMountain => $r };
                } elsif ($msg =~ m@^(VAL((?:$re_vall)+)) (.*)@o) {
                    $msg = $3;
                    $r->{s} = $1;
                    $r->{condVall} = ();
                    for ($2 =~ m@($re_vall)@og) {
                        my $m;
                        m@($re_phen_locdirs)? ($re_cond_vall)(?: ($re_chg_vall))?@o;

                        $m->{locationAndList} =_parseLocations $1 if defined $1;
                        $m->{condVallType}    = $2;
                        $m->{condVallChange}  = $3 if defined $3;
                        $m->{condVallType} =~ s/ /_/g;
                        $m->{condVallChange} =~ s/ /_/g
                            if exists $m->{condVallChange};
                        push @{$r->{condVall}}, $m;
                    }
                    push @{$metar{remark}}, { conditionValley => $r };
                } elsif ($msg =~ m@^(QU(L|K) ?([0-9/])(?: ?($re_compass_dir16))?) (.*)@o)
                {
                    $msg = $5;
                    $ft->{s}   = $1;
                    $ft->{txt} = $2 eq 'K'
                            ? "Condition of the sea:\t" . $sea_cond{$3}
                            : "Swell of the sea:\t" . $sea_swell{$3};
                    $ft->{txt} .= ' from the ' . $phen_loc{$4} if defined $4;
                } elsif ($msg =~ m@^(VIS (MAR) ([0-9]+) KM) (.*)@o) {
                    $msg = $4;
                    $r->{s} = $1;
                    $r->{locationAt} = $2;
                    $r->{visibility}{distance} = $3;
                    $r->{visibility}{unitLength} = 'KM';
                    push @{$metar{remark}}, { visAt => $r };
                } elsif ($msg =~ m@^((?:WIND THR|WT)($re_rwy_des) ($re_wind)) (.*)@o)
                {
                    $msg = $8;
                    push @{$metar{remark}}, { thrWind =>
                                              { s        => $1,
                                                rwyDesig => $2,
                                                wind     => _parseWind $3
                                              }
                                            };
                } elsif ($msg =~ m@^(VIS MIN ($re_vis_m)($re_compass_dir)?) (.*)@o) {
                    $msg = $4;
                    $r->{s}          = $1;
                    $r->{distance}   = $2 + 0;
                    $r->{unitLength} = 'M';
                    $r->{compassDir} = $3 if defined $3;
                    push @{$metar{remark}}, { visMin => $r };
                } else {
                    $parsed = 0;
                }
            } elsif ($cy[1] eq 'LK') {
                if ($msg =~ m@^(REG QNH ([01][0-9]{3})) (.*)@) {
                    $msg = $3;
                    push @{$metar{remark}}, { regQNH => {
                        s => $1,
                        hPa => $2
                    }};
                } else {
                    $parsed = 0;
                }
            } else {
                $parsed = 0;
            }
            if (!$parsed && _isValidCountry('remarkC_K_P_NS_TI_TJ_ET', \@cy))
            {
                $parsed = 1;
                if ($msg =~ m@^(4([01][0-9]{3})([01][0-9]{3})?) (.*)@o) {
                    $msg = $4;
                    $r->{s} = $1;
                    $r->{temp24hMax}{temp} = _parseUSTemp $2;
                    $r->{temp24hMax}{tempUnit} = 'C';
                    if (defined $3) {
                        $r->{temp24hMin}{temp} = _parseUSTemp $3;
                        $r->{temp24hMin}{tempUnit} = 'C';
                    }
                    push @{$metar{remark}}, { temp24h => $r };
                } elsif ($msg =~ m@^(4/([0-9]{3})) (.*)@o) {
                    $msg = $3;
                    $ft->{s}   = $1;
                    $ft->{txt} = "Snow on ground:\t\t" . $2 * 1 . " in.";
                } elsif ($msg =~ m@^((P|6|7)(?:([0-9]{4})|////)) (.*)@o) {
                    # Air Force Manual 15-111, Dec. '03, 2.8.1
                    $msg = $4;
                    $r->{s} = $1;
                    if (defined $3) {
                        $r->{precipVal} = $3 / 100;
                    } else {
                        $r->{notAvailable} = 1;
                    }
                    if ($2 eq 'P') {
                        $r->{precipHours} = 1;
                    } elsif ($2 eq '7') {
                        $r->{precipHours} = 24;
                    } else {
                        # EXTENSION: allow 20 minutes
                        if (exists $metar{obsTime}{hour}) {
                            if ((  $metar{obsTime}{hour}
                                 . $metar{obsTime}{minute}) =~
                             '(?:(?:23|05|11|17)[45]|(?:00|06|12|18)[01])[0-9]')
                            {
                                $r->{precipHours} = 6;
                            } elsif (  ($metar{obsTime}{hour}
                                     . $metar{obsTime}{minute}) =~
                             '(?:(?:02|08|14|20)[45]|(?:03|09|15|21)[01])[0-9]')
                            {
                                $r->{precipHours} = 3;
                            } else {
                                $r->{precipHoursNotAvailable} = 1;
                            }
                        } else {
                            $r->{precipHoursNotAvailable} = 1;
                        }
                    }
                    push @{$metar{remark}}, { precipHourly => $r };
                } elsif ($msg =~ m@^(933([0-9]{3})) (.*)@o) {
                    $msg = $3;
                    $ft->{s}   = $1;
                    $ft->{txt} = "Water equivalent of snow on ground: "
                                 . $2 / 10 . " in.";
                } elsif ($msg =~ m@^(CIG ([0-9]{3}V[0-9]{3})) (.*)@o) {
                    my $ceil = $2;
                    $msg = $3;

                    $ft->{s} = $1;
                    $ceil =~ /(...).(...)/;
                    $ft->{txt} = "Variable ceiling from:\t" . $1 * 100 . " ft\t\t\t\t\t"
                            . rnd($1 * 100 * $FT2M, 10) . " m\n"
                            . "\t\t   to:\t" . $2 * 100 . " ft\t\t\t\t\t"
                            . rnd($2 * 100 * $FT2M, 10) . " m";
                } elsif ($msg =~ m@^(($re_be_prec+)($re_phen_locdirs)?$re_wx_mov_d?) (.*)@o)
                {
                    $msg = $6;

                    $r->{beginEndPrecip}{s} = $1;
                    $r->{beginEndPrecip}{locationAndList} = _parseLocations $3
                        if defined $3;
                    $r->{beginEndPrecip}{$4} = _parseLocations $5 if defined $5;
                    $r->{beginEndPrecip}{precip} = ();
                    for ($2 =~ /($re_be_prec)+?/g) {
                        my @start_end;
                        my ($weather, $times) = /(.*?)($re_be_prec_be+)/g;
                        for ($times =~ /$re_be_prec_be+?/g) {
                            my %s_e;
                            /(.)(..)?(..)/;
                            $s_e{hour} = $2 if defined $2;
                            $s_e{minute} = $3;
                            if ($1 eq 'B') {
                                push @start_end, { beginPrecip => \%s_e };
                            } else {
                                push @start_end, { endPrecip => \%s_e };
                            }
                        }
                        push @{$r->{beginEndPrecip}{precip}},
                             { weather => _parseWeather($weather),
                               start_end => \@start_end
                             }
                    }
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ m@^(TR (SC|CI|AC)) (.*)@o) {
                    $msg = $3;
                    $ft->{s}   = $1;
                    $ft->{txt} = "Towering cloud:\t\t" . $cloud_types{$2};
                } elsif ($msg =~ m@^((SC|CI|AC) TR) (.*)@o) {
                    $msg = $3;
                    $ft->{s}   = $1;
                    $ft->{txt} = "Towering cloud:\t\t" . $cloud_types{$2};
                } elsif ($msg =~ m@^WSHFT (($re_hour)?($re_min)( FROPA)?) (.*)@o)
                {
                    $msg = $5;
                    $r->{windShift}{s} = "WSHFT $1";
                    $r->{windShift}{hour} = $2 if defined $2;
                    $r->{windShift}{minute} = $3;
                    $r->{windShift}{FROPA} = 1 if defined $4;
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ m@^((?:(THN) )?($re_cloud_cov) (ABV) ($re_cloud_base)) (.*)@o)
                {
                    $msg = $6;
                    $r->{cloudAbove}{s} = $1;
                    $r->{cloudAbove}{isThin} = 1 if defined $2;
                    $r->{cloudAbove}{cloudCover} = $3;
                    $r->{cloudAbove}{cloudBase} = $5 + 0;
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ m@^VIS ($re_vis_sm) ($re_compass_dir16) (.*)@o)
                {
                    my (@arr, $v);

                    $msg = $3;
                    $r->{visListLoc}{s} = "VIS $1 $2";
                    $r->{visListLoc}{arr} = \@arr;
                    $v->{locationAndList} = _parseLocations $2;
                    $v->{visibility}{distance} = _parseSM $1;
                    $v->{visibility}{unitLength} = 'SM';
                    push @arr, $v;
                    push @{$metar{remark}}, $r;
                } elsif ($msg =~ m@^VIS($re_phen_locdirs $re_vis_sm(?: SM)?(?:(?: AND)?$re_phen_locdirs $re_vis_sm(?: SM)?)*) (.*)@o)
                {
                    my @arr;

                    $msg = $2;
                    $r->{visListLoc}{s} = "VIS$1";
                    $r->{visListLoc}{arr} = \@arr;
                    for ($1 =~ m@(?:AND)?($re_phen_locdirs $re_vis_sm)@og) {
                        my $v;

                        m@($re_phen_locdirs) (\S+)@o;
                        $v->{locationAndList} = _parseLocations $1;
                        $v->{visibility}{distance} = _parseSM $2;
                        $v->{visibility}{unitLength} = 'SM';
                        push @arr, $v;
                    }
                    push @{$metar{remark}}, $r;
                } else {
                    $parsed = 0;
                }
            }
            if (!$parsed) {
                $msg =~ '(\S+) (.*)';
                $notRecognized .= ' ' unless $notRecognized eq '';
                $notRecognized .= $1;
                $msg = $2;
            }
            if ($parsed && $notRecognized ne '') {
                my $top = pop @{$metar{remark}};
                push @{$metar{remark}},
                         { freetext => { s   => $notRecognized,
                                         txt => _parseNotRec($notRecognized) }};
                push @{$metar{remark}}, $top;
                $notRecognized = '';
            }
            push @{$metar{remark}}, { freetext => $ft } if exists $ft->{txt};
        }
        push @{$metar{remark}}, 
                           { freetext => { s => $notRecognized,
                                           txt => _parseNotRec($notRecognized)}}
            if $notRecognized ne '';
    }

    if (   $is_taf
        && $msg =~ m@^(AMD (?:LTD TO CLD VIS AND WIND(?: (?:TIL ($re_hour|24)Z|(${re_hour})Z-($re_hour|24)Z))?|(NOT SKED))) (.*)@o)
    {
        $msg = $6;
        $metar{amendment}{s} = $1;
        if (defined $5) {
            $metar{amendment}{isNotScheduled} = 1;
        } else {
            $metar{amendment}{isLtdToCldVisWind} = 1;
            $metar{amendment}{hourTill} = $2 if defined $2;
            $metar{amendment}{hourFrom} = $3 if defined $3;
            $metar{amendment}{hourTill} = $4 if defined $4;
        }
    }

    if ($msg ne '') {
        $metar{ERROR}{descr} = 'other';
        $metar{ERROR}{pos}   = substr($metar{msgCorr}, 0,
                                 (length $metar{msgCorr}) - length $msg)
                        . '<@> ' . $msg;
    }
    return %metar;
}

=pod

=back

=cut

1;
__END__
